import h5py
import numpy as np


def _clean(value):
    """ Convert numpy numeric types to their python equivalents. """
    if isinstance(value, np.ndarray):
        if value.dtype.kind == 'S':
            return np.char.decode(value).tolist()
        else:
            return value.tolist()
    elif type(value).__module__ == np.__name__:
        # h5py==2.8.0 on windows sometimes fails to cast this from an np.float64 to a python.float
        # We have to let the user do this themselves, since casting here could be dangerous
        # https://github.com/h5py/h5py/issues/1051
        conversion = value.item()  # np.asscalar(value) was deprecated in v1.16
        if isinstance(conversion, bytes):
            conversion = conversion.decode()
        return conversion
    elif isinstance(value, bytes):
        return value.decode()
    else:
        return value


def _sanitize_data_for_writing(data):
    # To make the interface more user friendly we encode python strings as  byte-strings when writing datasets
    if isinstance(data, str):
        # Plain python-strings can be encoded trivially
        return data.encode()
    elif isinstance(data, np.ndarray) and data.dtype.kind == np.dtype(np.unicode_):
        # If the array is all of one type, unicode-string, we can encode with numpy
        return data.astype('S')
    elif isinstance(data, np.ndarray) and len(data.dtype) > 1:
        # If the array is of mixed types we have to set the encoding column by column
        encoded_dtypes = []
        for field_name in data.dtype.names:
            field_dtype, field_byte_index = data.dtype.fields[field_name]
            if field_dtype.kind == 'U':
                str_len = field_dtype.itemsize // field_dtype.alignment
                field_dtype = np.dtype("|S{}".format(str_len))
            encoded_dtypes.append((field_name, field_dtype))
        return data.astype(encoded_dtypes)

    return data


def _sanitize_data_for_reading(data):
    # To make the interface more user friendly we decode byte-strings into unicode strings when reading datasets
    if isinstance(data, h5py.Dataset):
        data = data[()]

    if isinstance(data, bytes):
        # Plain byte-strings can be decoded trivially
        return data.decode()
    elif isinstance(data, np.ndarray) and data.dtype.kind == 'S':
        # If the array is all of one type, byte-string, we can decode with numpy
        return np.char.decode(data)
    elif isinstance(data, np.ndarray) and len(data.dtype) > 1:
        # If the array is of mixed types we have to decode column by column
        decoded_dtypes = []
        for field_name in data.dtype.names:
            field_dtype, field_byte_index = data.dtype.fields[field_name]
            if field_dtype.kind == 'S':
                field_dtype = np.dtype("<U{}".format(field_dtype.itemsize))
            decoded_dtypes.append((field_name, field_dtype))
        return data.astype(decoded_dtypes)

    return data



from pathlib import Path
from test.helpers import TestFast5ApiHelper, test_data
from unittest.mock import patch
from ont_fast5_api.conversion_tools.demux_fast5 import Fast5Demux
from ont_fast5_api.multi_fast5 import MultiFast5File


class TestDemuxFast5(TestFast5ApiHelper):
    multi_fast5 = Path(test_data) / "multi_read" / "batch_0.fast5"
    summary = Path(test_data) / "summaries" / "two_barcode_summary.txt"
    barcode01 = {"fe85b517-62ee-4a33-8767-41cab5d5ab39", "fe8a3026-d1f4-46b3-8daa-e610f27acde1"}
    barcode02 = {"fe9374ee-b86a-4ca4-81dc-ac06e3297728", "fe849dd3-63bc-4044-8910-14e1686273bb"}
    barcodes = (barcode01, barcode02)

    @patch('ont_fast5_api.conversion_tools.demux_fast5.logging')
    @patch('ont_fast5_api.conversion_tools.conversion_utils.ProgressBar')
    def test_demux_1t(self, mock_pbar, mock_logger):
        # given 4 read multi fast5 file and a summary, bin it in two barcode directories
        output_dir = Path(self.save_path) / "1t"
        output_dir.mkdir()
        demux = Fast5Demux(input_dir=self.multi_fast5, output_dir=output_dir, summary_file=self.summary,
                           demultiplex_column="barcode_arrangement",threads=1)
        demux.run_batch()
        self.check_output(output_dir)

    @patch('ont_fast5_api.conversion_tools.demux_fast5.logging')
    @patch('ont_fast5_api.conversion_tools.conversion_utils.ProgressBar')
    def test_demux_8t(self, mock_pbar, mock_logger):
        # given 4 read multi fast5 file and a summary, bin it in two barcode directories
        output_dir = Path(self.save_path) / "8t"
        output_dir.mkdir()
        demux = Fast5Demux(input_dir=self.multi_fast5, output_dir=output_dir, summary_file=self.summary,
                           demultiplex_column="barcode_arrangement",threads=8)
        demux.workers_setup()
        # even with 8 threads allocated, only max 2 can be used
        self.assertEqual(demux.max_threads, 2)
        demux.run_batch()
        self.check_output(output_dir)

    def check_output(self, result_path):
        output_dir1 = result_path / "barcode01"
        output_dir2 = result_path / "barcode02"

        for directory, barcodes in zip((output_dir1, output_dir2), self.barcodes):
            self.assertTrue(directory.exists())
            self.assertTrue(directory.is_dir())
            batch_file = directory / "batch0.fast5"
            self.assertTrue(batch_file.exists())
            self.assertTrue(batch_file.is_file())
            with MultiFast5File(batch_file, 'r') as fast5_in:
                read_ids = set(fast5_in.get_read_ids())
                self.assertEqual(read_ids, barcodes)
            summary_file = directory / "filename_mapping.txt"
            self.assertTrue(summary_file.exists())
            self.assertTrue(summary_file.is_file())

    @patch('ont_fast5_api.conversion_tools.demux_fast5.logging')
    @patch('ont_fast5_api.conversion_tools.conversion_utils.ProgressBar')
    def test_parse_summary(self, mock_pbar, mock_logger):
        # create a summary file with standard column names
        summary_file = Path(self.generate_temp_filename())
        truth = {"barcode01": self.barcode01, "barcode02": self.barcode02}
        with open(summary_file, 'w') as summ:
            header = "read_id\tbarcode_arrangement\n"
            summ.write(header)
            for barcode, read_ids in truth.items():
                for read_id in read_ids:
                    line = read_id + "\t" + barcode + "\n"
                    summ.write(line)

        demux = Fast5Demux(input_dir=self.multi_fast5, output_dir=Path(self.save_path), summary_file=summary_file,
                           demultiplex_column="barcode_arrangement")
        demux.workers_setup()

        self.assertEqual(demux.read_sets, truth)

        # create a summary file with non-standard column names
        truth = {"genome1": self.barcode01, "genome2": self.barcode02}
        summary_file = Path(self.generate_temp_filename())
        with open(summary_file, 'w') as summ:
            header = "genome\tread_name\n"
            summ.write(header)
            for genome, read_ids in truth.items():
                for read_id in read_ids:
                    line = genome + "\t" + read_id + "\n"
                    summ.write(line)

        demux = Fast5Demux(input_dir=self.multi_fast5, output_dir=Path(self.save_path), summary_file=summary_file,
                           demultiplex_column="genome", read_id_column="read_name")
        demux.workers_setup()

        self.assertEqual(demux.read_sets, truth)



import os
import numpy as np
from ont_fast5_api.fast5_file import Fast5File
from ont_fast5_api.analysis_tools.event_detection import EventDetectionTools
from test.helpers import TestFast5ApiHelper, test_data


class TestEventDetectionTools(TestFast5ApiHelper):

    def test_001_read_events(self):
        # Check that it is recognized properly.
        fname = os.path.join(test_data, 'read_file_v1.0_single.fast5')
        with EventDetectionTools(fname, mode='r', ) as fh:
            self.assertTrue(fh.has_event_data)
            self.assertTrue(fh.has_event_data(read_number=59))
            self.assertEqual('EventDetection_000', fh.group_name)
            data, attrs = fh.get_event_data()
            self.assertDictEqual({'read_number': 59,
                                  'strand_id': 60,
                                  'start_mux': 1,
                                  'end_mux': 1,
                                  'start_time': 32463855,
                                  'duration': 729468}, attrs)
            self.assertEqual(7875, data.size)
            self.assertEqual(118, data[0]['length'])
            data, attrs = fh.get_event_data(time_in_seconds=True)
            self.assertEqual(0.0236, data[0]['length'])

    def test_002_write_events(self):
        fname = os.path.join(self.save_path, 'test.fast5')
        with Fast5File(fname, 'w') as fh:
            fh.add_channel_info({'channel_number': 1, 'sampling_rate': 4000})
            fh.add_read(12, 'unique_snowflake', 12345, 111, 0, 120.75)
            with EventDetectionTools(fh, group_name='EventDetection_000',
                                     meta={'name': 'test', 'version': '0.1.0'}) as evdet:
                data = np.zeros(100, dtype=[('start', int), ('length', int), ('mean', float), ('stdv', float)])
                read_attrs = {'read_number': 12}
                evdet.set_event_data(data, read_attrs)
        with Fast5File(fname, 'r') as fh:
            self.assertEqual(1, len(fh.status.read_info))
            read_info = fh.status.read_info[0]
            self.assertEqual(12, read_info.read_number)
            group = fh.get_latest_analysis('EventDetection')
            self.assertEqual('EventDetection_000', group)
            with EventDetectionTools(fh) as evdet:
                self.assertTrue(evdet.has_event_data())
                data, attrs = evdet.get_event_data()
                self.assertDictEqual({u'read_number': 12,
                                      u'read_id': 'unique_snowflake',
                                      u'start_time': 12345,
                                      u'duration': 111,
                                      u'start_mux': 0,
                                      u'median_before': 120.75}, attrs)
                self.assertEqual(100, data.size)




from ont_fast5_api.fast5_file import Fast5File, Fast5FileTypeError
from ont_fast5_api.fast5_info import _clean
from ont_fast5_api.multi_fast5 import MultiFast5File

MULTI_READ = "multi-read"
SINGLE_READ = "single-read"
BULK_FAST5 = "bulk"


def get_fast5_file(filepath, mode='r', driver=None):
    if is_multi_read(filepath):
        return MultiFast5File(filepath, mode, driver=driver)
    else:
        return Fast5File(filepath, mode)


def check_file_type(f5_file):
    try:
        return _clean(f5_file.handle.attrs['file_type'])
    except KeyError:
        # On older files we don't have the 'file_type' attribute so check groups explicitly
        if len(f5_file.handle) == 0:
            # If there are no top-level groups we default to MultiRead
            return MULTI_READ
        if len([read for read in f5_file.handle if read.startswith('read_')]) != 0:
            # If there are any read_xxx groups we're definitely MultiRead
            return MULTI_READ
        if "UniqueGlobalKey" in f5_file.handle:
            # This group indicates a single read
            return SINGLE_READ
    raise TypeError("Fast5 file type could not be identified as single- or multi-read. "
                    "\nFilepath: {}".format(f5_file.filename))


def is_multi_read(filepath):
    """
    Determine if a file is a MultiFast5File, True if it is, False if it is a single Fast5File error for other types
    """
    with MultiFast5File(filepath, mode='r') as f5_file:
        file_type = check_file_type(f5_file)
        if file_type == MULTI_READ:
            return True
        elif file_type == SINGLE_READ:
            return False
        elif file_type == BULK_FAST5:
            raise NotImplementedError("ont_fast5_api does not support bulk fast files: {}".format(filepath))
        raise Fast5FileTypeError("Unknown file type: '{}' for file: {}".format(file_type, filepath))



"""Filter Fast5 files based on read_id list
"""
import csv
import logging
from argparse import ArgumentParser
from math import ceil
from multiprocessing import Pool
from os import makedirs, path
from pathlib import Path
from time import sleep

from ont_fast5_api.compression_settings import COMPRESSION_MAP
from ont_fast5_api.conversion_tools.conversion_utils import get_fast5_file_list, get_progress_bar, Fast5FilterWorker
from ont_fast5_api.conversion_tools.conversion_utils import READS_PER_FILE, FILENAME_BASE

logging.basicConfig(level=logging.DEBUG)


class Fast5Filter:
    """
    Extract reads listed read_list_file from fast5 files in input_folder, write to multi-fast5 files in
    output_folder
    """

    def __init__(self, input_folder, output_folder, read_list_file, filename_base=FILENAME_BASE,
                 batch_size=READS_PER_FILE, threads=1, recursive=False, file_list_file=None, follow_symlinks=True,
                 target_compression=None):
        assert path.isdir(input_folder)
        assert path.isfile(read_list_file)
        assert isinstance(filename_base, str)
        assert isinstance(batch_size, int)
        assert isinstance(threads, int)
        assert isinstance(recursive, bool)
        self.logger = logging.getLogger(self.__class__.__name__)

        self.read_set = parse_summary_file(read_list_file)
        self.input_f5s = get_fast5_file_list(str(input_folder), recursive, follow_symlinks=follow_symlinks)
        makedirs(output_folder, exist_ok=True)

        if len(self.read_set) < 1:
            raise ValueError("No reads in read list file {}".format(read_list_file))

        if len(self.input_f5s) < 1:
            raise ValueError(
                "No input fast5 files found in {}. Recursion is set to {}".format(str(input_folder), recursive))

        if batch_size < 1:
            raise ValueError("Batch size (--batch_size) must be a positive integer, not {}".format(batch_size))

        if threads < 1:
            raise ValueError("Max number of threads (--threads) must be a positive integer, not {}".format(threads))

        if file_list_file:
            file_set = parse_summary_file(file_list_file)
            for file in file_set:
                assert path.exists(file), "{} from file list doesn't exist".format(file)
            self.input_f5s = list(file_set.intersection(self.input_f5s))

        # determine max number of workers
        num_outputs = int(ceil(len(self.read_set) / float(batch_size)))
        self.num_workers = min(threads, num_outputs, len(self.input_f5s))

        # progressbar total is number of reads in read_set plus number of input files
        # (to see progress while scanning files that don't have any relevant reads)
        self.pbar = get_progress_bar(len(self.read_set) + len(self.input_f5s))

        self.worker = Fast5FilterWorker(
            input_file_list=self.input_f5s,
            output_dir=Path(output_folder),
            logger=self.logger,
            progressbar=self.pbar,
            read_set=self.read_set,
            filename_base=filename_base,
            batch_size=batch_size,
            target_compression=target_compression
        )

    def run_batch(self):

        if self.num_workers == 1:
            self.worker.run_batch(pool=None)
        else:
            with Pool(self.num_workers) as pool:
                self.worker.run_batch(pool=pool)

                while self.worker.tasks:
                    sleep(1)

                pool.close()
                pool.join()

        self.pbar.finish()
        self.logger.info("{} reads extracted".format(sum(len(v) for v in self.worker.out_files.values())))

        # report reads not found
        if len(self.worker.read_set) > 0:
            self.logger.warning("{} reads not found!".format(len(self.worker.read_set)))


def parse_summary_file(read_list_file):
    """
    Opens a text file and returns set of read_ids
    Expects either a single column file where every line is read_id or
      multi-column Tab-separated CSV, that contains a column read_id
    :param read_list_file: path to file
    :return: set
    """
    reads = set()
    with open(str(read_list_file), 'r') as fh:
        read_list_tsv = csv.reader(fh, delimiter='\t')
        header = next(read_list_tsv)

        if "read_id" in header:
            col_idx = header.index("read_id")
        else:
            if len(header) == 1:
                reads.add(header[0].strip())
                col_idx = 0
            else:
                raise TypeError("multi-column file without 'read_id' column")

        for line in read_list_tsv:
            reads.add(line[col_idx].strip())
    if len(reads) < 1:
        raise ValueError("No reads in read list file {}".format(read_list_file))
    return reads


def main():
    parser = ArgumentParser("Tool for extracting reads from a multi_read_fast5_file by read_id")
    parser.add_argument('-i', '--input', required=True,
                        help="Path to Fast5 file or directory of Fast5 files")
    parser.add_argument('-s', '--save_path', required=True,
                        help="Folder to output MultiRead subset to")
    parser.add_argument('-l', '--read_id_list', required=True,
                        help="File containing list of read ids to extract (or sequencing_summary.txt file)")
    parser.add_argument('-f', '--filename_base', default=FILENAME_BASE, required=False,
                        help="Root of output filename, default='{}' -> '{}0.fast5'".format(FILENAME_BASE, FILENAME_BASE))
    parser.add_argument('-n', '--batch_size', type=int, default=READS_PER_FILE, required=False,
                        help="Number of reads per multi-read file (default {}".format(READS_PER_FILE))
    parser.add_argument('-t', '--threads', type=int, default=1, required=False,
                        help="Maximum number of threads to use")
    parser.add_argument('-r', '--recursive', action='store_true', required=False, default=False,
                        help="Search recursively through folders for MultiRead fast5 files")
    parser.add_argument('--ignore_symlinks', action='store_true',
                        help="Ignore symlinks when searching recursively for fast5 files")
    parser.add_argument('-c', '--compression', required=False, default=None,
                        choices=list(COMPRESSION_MAP.keys()) + [None], help="Target output compression type")
    parser.add_argument('--file_list', required=False,
                        help="File containing names of files to search in")
    args = parser.parse_args()

    if args.compression is not None:
        args.compression = COMPRESSION_MAP[args.compression]

    multifilter = Fast5Filter(input_folder=args.input,
                              output_folder=args.save_path,
                              filename_base=args.filename_base,
                              read_list_file=args.read_id_list,
                              batch_size=args.batch_size,
                              threads=args.threads,
                              recursive=args.recursive,
                              file_list_file=args.file_list,
                              follow_symlinks=not args.ignore_symlinks,
                              target_compression=args.compression)

    multifilter.run_batch()


if __name__ == '__main__':
    main()



import os

from ont_fast5_api.compression_settings import VBZ
from ont_fast5_api.conversion_tools.compress_fast5 import compress_file
from ont_fast5_api.conversion_tools.conversion_utils import extract_selected_reads
from ont_fast5_api.fast5_interface import get_fast5_file
from ont_fast5_api.multi_fast5 import MultiFast5File
from ont_fast5_api.static_data import HARDLINK_GROUPS
from test.helpers import TestFast5ApiHelper, test_data


class TestHardlinkMetaData(TestFast5ApiHelper):
    read_subset = {'00031f3e-415c-4ab5-9c16-fb6fe45ff519',
                   "000c0b4e-46c2-4fb5-9b17-d7031eefb975",
                   '000ebd63-3e1a-4499-9ded-26af3225a022',
                   '002ad0e4-c6bb-4eff-a30f-5fec01475ab8',
                   '0059d270-3238-4413-b38b-f588e28326df'}

    def test_create_read(self):
        input_path = os.path.join(test_data, 'hardlink', 'unlinked', 'batch0.fast5')
        output_path = self.generate_temp_filename()
        compress_file(input_path, output_path, target_compression=VBZ)
        new_read_id = "123456789abcdef"
        with MultiFast5File(output_path, 'a') as f5:
            # Test we can hardlink to existing metadata when creating an new empty read
            run_id = list(f5.run_id_map.keys())[0]
            master_read_id = f5.run_id_map[run_id]
            f5.create_empty_read(new_read_id, run_id)
            for group in HARDLINK_GROUPS:
                self.assertTrue(self.is_read_hardlinked(f5, new_read_id, master_read_id, group))

            # Test we don't explode if there is no metadata
            f5.create_empty_read(new_read_id[::-1], "not an existing run_id")

    def test_hardlink_multi_compression(self):
        input_path = os.path.join(test_data, 'hardlink', 'unlinked', 'batch0.fast5')
        output_path = self.generate_temp_filename()

        self.assertFalse(self.is_file_hardlinked(input_path))
        compress_file(input_path, output_path, target_compression=VBZ)
        self.assertTrue(self.is_file_hardlinked(output_path))

    def test_hardlink_subset(self):
        input_path = os.path.join(test_data, 'hardlink', 'unlinked', 'batch0.fast5')
        output_path = self.generate_temp_filename()

        self.assertFalse(self.is_file_hardlinked(input_path))
        extract_selected_reads(input_path, output_path, self.read_subset, count=len(self.read_subset))
        self.assertTrue(self.is_file_hardlinked(output_path))

    def test_hardlink_subset_single_reads(self):
        input_path = os.path.join(test_data, 'hardlink', 'single_reads')
        output_path = self.generate_temp_filename()

        for single_read_file in os.listdir(input_path):
            extract_selected_reads(os.path.join(input_path, single_read_file), output_path, self.read_subset, count=1)
        self.assertTrue(self.is_file_hardlinked(output_path))

    def test_hardlink_single_to_multi(self):
        input_folder = os.path.join(test_data, 'hardlink', 'single_reads')
        input_files = [os.path.join(input_folder, file) for file in os.listdir(input_folder)]
        output_path = self.generate_temp_filename()

        with MultiFast5File(output_path, 'a') as multi_f5:
            for input_file in input_files:
                with get_fast5_file(input_file, 'r') as f5_file:
                    for read in f5_file.get_reads():
                        multi_f5.add_existing_read(read)

        with MultiFast5File(output_path, 'r') as multi_f5:
            self.assertEqual(len(input_files), len(multi_f5.get_read_ids()))
        self.assertTrue(self.is_file_hardlinked(output_path))

    def is_file_hardlinked(self, input_path):
        file_hardlinked = True
        with MultiFast5File(input_path, 'r') as f5_file:
            for read in f5_file.get_reads():
                master_read_id = f5_file.run_id_map[read.get_run_id()]
                for group in HARDLINK_GROUPS:
                    file_hardlinked &= self.is_read_hardlinked(f5_file, read.read_id, master_read_id, group)
        return file_hardlinked

    def is_read_hardlinked(self, f5_handle, read_id1, read_id2, group):
        if read_id1 == read_id2:
            return True
        group1 = f5_handle.get_read(read_id1).handle[group]
        group2 = f5_handle.get_read(read_id2).handle[group]
        return group1 == group2



import logging
import os
from tempfile import TemporaryDirectory, _get_candidate_names
import unittest

test_data = os.path.join(os.path.dirname(__file__), 'data')


def disable_logging(test_func):
    def do_test(self, *args, **kwargs):
        logging.disable(logging.CRITICAL)
        test_func(self, *args, **kwargs)

    return do_test


class TestFast5ApiHelper(unittest.TestCase):

    def setUp(self):
        self._tmp_dir = TemporaryDirectory()
        self.save_path = self._tmp_dir.name

    def tearDown(self):
        self._tmp_dir.cleanup()

    def generate_temp_filename(self):
        return os.path.join(self.save_path, next(_get_candidate_names()))



""" Basic API for reading/writing single-read fast5 files.
"""
import os
import h5py
import numpy as np

from ont_fast5_api import CURRENT_FAST5_VERSION
from ont_fast5_api.compression_settings import VBZ, raise_missing_vbz_error_read
from ont_fast5_api.fast5_read import Fast5Read
from ont_fast5_api.fast5_info import Fast5Info, ReadInfo
from ont_fast5_api.static_data import supported_modes, mode_docstring

# This unused import is included for backwards compatibilty and can be removed in future.
from ont_fast5_api.data_sanitisation import _sanitize_data_for_reading, _sanitize_data_for_writing


class Fast5FileTypeError(ValueError):
    pass


class Fast5File(Fast5Read):
    """ This object encapsulates a read fast5 file. It can be used
    instead of directly using the h5py API in order to help maintain
    consistency in fast5 file format, and simplify common tasks.

    The object will contain a field called **status** that is a
    Fast5Status object with details about the file.
    """

    def __init__(self, fname, mode='r'):
        """ Constructor. Opens the specified file.

        :param fname: Filename to open.
        :param mode: File open mode (r, r+, w, w-, x, a).
        """
        self.global_key = "UniqueGlobalKey/"
        if mode not in supported_modes:
            raise IOError("Unsupported file handle mode : '{}' {}".format(mode, mode_docstring))
        self.filename = fname
        self.handle = None
        self.mode = mode
        self._initialise_file()

    def get_reads(self):
        # We yield here for consistency with MultiFast5File
        yield self.get_read(self.get_read_id())

    def get_read_ids(self):
        return [self.get_read_id()]

    def get_read(self, read_id):
        if read_id != self.get_read_id():
            raise KeyError("read_id given: {} does not match read_id in file: {}"
                           "".format(read_id, self.get_read_id()))
        return self

    @property
    def read_id(self):
        return self.status.read_info[0].read_id

    def get_read_id(self):
        DeprecationWarning("'read.get_read_id()' will be deprecated. Use the property 'read.read_id' instead")
        return self.read_id

    @property
    def raw_dataset_group_name(self):
        read_number = self._get_only_read_number()
        return 'Raw/Reads/Read_{}'.format(read_number)

    @property
    def has_context_tags(self):
        return 'context_tags' in self.handle[self.global_key[:-1]]

    def add_channel_info(self, data, clear=False):
        """ Add channel info data to the channel_id group.

        :param data: A dictionary of key/value pairs. Keys must be strings.
            Values can be strings or numeric values.
        :param clear: If set, any existing channel info data will be removed. 
        """
        self.assert_writeable()
        self._add_attributes(self.global_key + 'channel_id', data, clear)

    def get_raw_data(self, read_number=None, start=None, end=None, scale=False):
        """ Pull raw data from the file.
        
        :param read_number: The read number you want raw data from. Pass None
            if there is only one read, and will grab that read's raw data.
        :param start: The first sample to pull. Default is to pull from the
            beginning.
        :param end: One past the last sample to pull. Default is to pull until
            the end.
        :param scale: If set, the returned data will be scaled floating point
            values, in pA. Otherwise raw DAQ values are returned as 16 bit
            integers.
        :returns: Raw data as either 32 bit floats, or 16 bit integers.
        """
        self.assert_open()
        if self.raw_dataset_name not in self.handle:
            msg = 'Fast5 file has no raw data for read {} in {}'.format(read_number, self.filename)
            raise KeyError(msg)
        return self._load_raw(self.raw_dataset_name, start, end, scale)

    def add_raw_data(self, data, attrs=None, compression=VBZ):
        """ Add raw data for a read.
        
        :param read_number: The number of the read the raw data is for.
        :param data: The raw data DAQ values (16 bit integers).
        
        The read must already exist in the file. It must not already
        have raw data.
        """
        self.assert_writeable()

        if self.raw_dataset_name in self.handle:
            msg = 'Fast5 file already has raw data for read {} in {}'
            raise KeyError(msg.format(self.raw_dataset_name, self.filename))

        group_name = self.raw_dataset_group_name
        if group_name not in self.handle:
            self.handle.create_group(group_name)

        try:
            self.handle[group_name].create_dataset('Signal', data=data, dtype='i2', **vars(compression))
        except ValueError as e:
            raise_missing_vbz_error_read(e)

        read_index = self.status.read_number_map[self._get_only_read_number()]
        self.status.read_info[read_index].has_raw_data = True
        if attrs is not None:
            self._add_attributes(group_name, attrs, clear=True)

    def add_read(self, read_number, read_id, start_time, duration, mux, median_before):
        """ Add a new read to the file.

        :param read_number: The read number to assign to the read.
        :param read_id: The unique read-id for the read.
        :param start_time: The start time (in samples) of the read.
        :param duration: The duration (in samples) of the read.
        :param mux: The mux set at the time of the read.
        :param median_before: The median level of the data before the read.

        Note that most tools assume a file contains only one read.
        Putting multiple reads into a file severely limits the
        ability to operate on those reads with standard tools.
        """
        self.assert_writeable()
        read_info = ReadInfo(read_number, read_id, start_time, duration, mux=mux, median_before=median_before)
        self.status.read_info.append(read_info)
        n = len(self.status.read_info) - 1
        self.status.read_number_map[read_number] = n
        self.status.read_id_map[read_id] = n
        group_name = self.raw_dataset_group_name
        attrs = {'read_number': read_number,
                 'read_id': read_id,
                 'start_time': start_time,
                 'duration': duration,
                 'start_mux': mux,
                 'median_before': median_before}
        self._add_group(group_name, attrs)

    #########################
    #
    #  Static methods below
    #
    #########################

    @staticmethod
    def update_legacy_file(fname):
        """ Update a fast5 file from an older version to the new standard.
        
        :param fname: The filename of the fast5 file.
        """
        status = Fast5Info(fname)
        if not status.valid:
            raise IOError('Cannot update invalid file: {}'.format(fname))
        with h5py.File(fname, 'r+') as handle:
            # Add Raw/Read/Read_## groups for reads if they are missing.
            for read_info in status.read_info:
                read_group_name = 'Raw/Reads/Read_{}'.format(read_info.read_number)
                if read_group_name in handle:
                    rgh = handle[read_group_name]
                else:
                    rgh = handle.create_group(read_group_name)
                rgh.attrs['read_number'] = read_info.read_number
                rgh.attrs['read_id'] = read_info.read_id
                rgh.attrs['duration'] = read_info.duration
                rgh.attrs['start_time'] = read_info.start_time
                rgh.attrs['start_mux'] = read_info.start_mux

            # Add the Analyses and tracking_id groups, if they are missing.
            if not 'Analyses' in handle:
                handle.create_group('Analyses')
            if not 'tracking_id' in handle['UniqueGlobalKey']:
                handle.create_group('UniqueGlobalKey/tracking_id')

            # Update the EventDetection_000 created by old versions of MinKNOW, if needed.
            if status.version < 1.1:
                if 'Analyses/EventDetection_000' in handle:
                    reads_group = handle['Analyses/EventDetection_000/Reads']
                    data_group_names = reads_group.keys()
                    for data_group_name in data_group_names:
                        read_group = reads_group[data_group_name]
                        read_number = read_group.attrs['read_number']
                        read_info = status.read_info[status.read_number_map[read_number]]
                        read_group.attrs['read_id'] = read_info.read_id
                        if 'Events' in read_group:
                            dataset = read_group['Events']
                            if 'variance' in dataset.dtype.names:
                                old_data = read_group['Events'][()]
                                new_data = np.empty(old_data.size, dtype=[('mean', float), ('stdv', float),
                                                                          ('start', int), ('length', int)])
                                new_data[:]['mean'] = old_data['mean']
                                new_data[:]['stdv'] = np.sqrt(old_data['variance'])
                                new_data[:]['start'] = old_data['start']
                                new_data[:]['length'] = old_data['length']
                                del read_group['Events']
                                read_group.create_dataset('Events', data=new_data, compression='gzip')

            # Update the version number.
            handle.attrs['file_version'] = CURRENT_FAST5_VERSION

    @staticmethod
    def read_summary_data(fname, component):
        """ Read summary data suitable to encode as a json packet.

        :param fname: The fast5 file to pull the summary data from.
        :param component: The component name to pull summary data for.

        :returns: A dictionary containing the summary data.
        """
        summary = {}
        with Fast5File(fname, mode='r') as fh:
            summary['tracking_id'] = fh.get_tracking_id()
            summary['channel_id'] = fh.get_channel_info()
            read_info = fh.status.read_info
            read_summary = []
            for read in read_info:
                read_summary.append({'read_number': read.read_number,
                                     'read_id': read.read_id,
                                     'start_time': read.start_time,
                                     'duration': read.duration,
                                     'start_mux': read.start_mux})
            summary['reads'] = read_summary
            analyses_list = fh.list_analyses(component)
            _, group_names = zip(*analyses_list)
            group_names = sorted(group_names)
            group = group_names[-1]
            summary['software'] = fh.get_analysis_attributes(group)
            summary['software']['component'] = group[:-4]

            summary['data'] = fh.get_summary_data(group)
            summary['filename'] = os.path.basename(fname)
        return summary

    ##########################
    #
    #  Private methods below
    #
    ##########################

    def _get_only_read_number(self):
        read_number = self.status.read_info[0].read_number
        return read_number

    def _initialise_file(self):
        try:
            if self.mode in ['w', 'w-', 'x']:
                with h5py.File(self.filename, self.mode) as fh:
                    fh.attrs['file_version'] = CURRENT_FAST5_VERSION
                    fh.create_group('Analyses')
                    fh.create_group('Raw/Reads')
                    fh.create_group(self.global_key + 'channel_id')
                    fh.create_group(self.global_key + 'context_tags')
                    fh.create_group(self.global_key + 'tracking_id')
                self.mode = 'r+'
            self.status = Fast5Info(self.filename)
            if self.status.valid:
                self.handle = h5py.File(self.filename, self.mode)
        except Exception:
            raise Fast5FileTypeError("Failed to initialise single-read Fast5File: '{}'".format(self.filename))


class EmptyFast5(Fast5File):
    def _initialise_file(self):
        # Enable creation of Fast5File without metadata, which we will populate later
        self.handle = h5py.File(self.filename, self.mode)
        self.handle.attrs['file_version'] = CURRENT_FAST5_VERSION



import os
import unittest

from ont_fast5_api.fast5_file import Fast5File
from ont_fast5_api.fast5_interface import get_fast5_file, check_file_type, MULTI_READ, SINGLE_READ
from ont_fast5_api.multi_fast5 import MultiFast5File
from test.helpers import test_data



class TestFast5Interface(unittest.TestCase):

    def test_correct_type(self):
        single_read_path = os.path.join(test_data, "single_reads", "read0.fast5")
        single_read_id = Fast5File(single_read_path).get_read_id()
        with get_fast5_file(single_read_path) as f5:
            self.assertTrue(isinstance(f5, Fast5File))
            self.assertEqual(check_file_type(f5), SINGLE_READ)
            self.assertEqual(len(f5.get_read_ids()), 1)
            self.assertEqual(single_read_id, f5.get_read_ids()[0])
            self.get_raw(f5)

        multi_read_path = os.path.join(test_data, "multi_read", "batch_0.fast5")
        with get_fast5_file(multi_read_path) as f5:
            self.assertTrue(isinstance(f5, MultiFast5File))
            self.assertEqual(check_file_type(f5), MULTI_READ)
            self.assertTrue(len(f5.get_read_ids()) >= 1)
            self.get_raw(f5)

    def get_raw(self, f5):
        # Test we can get raw data using the same method for single and multi
        raw_data = f5.get_read(f5.get_read_ids()[0]).get_raw_data()
        self.assertTrue(len(raw_data) >= 0)



import multiprocessing
import os

from glob import glob
from math import ceil
from pathlib import Path
from typing import List, Optional, Set, Tuple, Iterator

from progressbar import RotatingMarker, ProgressBar, SimpleProgress, Bar, Percentage, ETA
from logging import Logger

from ont_fast5_api.fast5_interface import get_fast5_file
from ont_fast5_api.multi_fast5 import MultiFast5File
from ont_fast5_api.fast5_read import Fast5Read


READS_PER_FILE = 4000
FILENAME_BASE = "batch"


def batcher(iterable, n=1):
    l = len(iterable)
    for ndx in range(0, l, n):
        yield iterable[ndx:min(ndx + n, l)]


def yield_fast5_files(input_path, recursive, follow_symlinks=True):
    """
    Yield fast5 file paths within a given directory.
    Optionally search recursively and follow symbolic links

    :param input_path: Path
    :param recursive: bool
    :param follow_symlinks: bool
    :return:
    """
    if os.path.isfile(input_path):
        yield input_path
        return

    if recursive:
        for root, _, filenames in os.walk(input_path, followlinks=follow_symlinks):
            for filename in filenames:
                if not filename.endswith('.fast5'): continue
                yield os.path.join(root, filename)
    else:
        for filename in glob(os.path.join(input_path, '*.fast5')):
            yield filename
    return


def yield_fast5_reads(input_path, recursive, follow_symlinks=True, read_ids=None):
    """
    Iterate over reads in fast5 files and yield read_ids and fast5 read objects.
    If read_id_set is defined, skip reads which are not in this set/list. An empty set/list returns all.

    :param input_path: Path
    :param recursive: bool
    :param follow_symlinks: bool
    :param read_ids: set or list
    :raise TypeError: read_id_set must be of type set or list'
    :return: yielded tuple (read_id, fast5_read_object)
    """
    if not isinstance(read_ids, (list, set)) and read_ids is not None:
        raise TypeError('read_ids must be of type set or list or none')

    if read_ids and isinstance(read_ids, list):
        read_ids = set(read_ids)

    for fast5_path in yield_fast5_files(input_path=input_path, recursive=recursive, follow_symlinks=follow_symlinks):
        fast5_file = get_fast5_file(fast5_path)
        if read_ids:
            selected_reads = read_ids.intersection(fast5_file.get_read_ids())
        else:
            selected_reads = fast5_file.get_read_ids()

        for read_id in selected_reads:
            yield read_id, fast5_file.get_read(read_id)


def get_fast5_file_list(input_path, recursive, follow_symlinks=True):
    # NB this method is provided for compatibility with use cases where
    # generator behaviour is not appropriate.
    # E.g. where we need to know the file_list length or be able to sample from it
    return list(yield_fast5_files(input_path, recursive, follow_symlinks))


def get_progress_bar(num_reads):
    bar_format = [RotatingMarker(), " ", SimpleProgress(), Bar(), Percentage(), " ", ETA()]
    progress_bar = ProgressBar(maxval=num_reads, widgets=bar_format)
    bad_progressbar_version = False
    try:
        progress_bar.currval
    except AttributeError as e:
        bad_progressbar_version = True
        pass
    if bad_progressbar_version:
        raise RuntimeError('Wrong progressbar package detected, likely '
                           '"progressbar2". Please uninstall that package and '
                           'install "progressbar33" instead.')

    return progress_bar.start()


class Fast5FilterWorker:
    """
    Common worker for fast5_subset and demux_fast5
    Extract reads in [read_set] from fast5 files in [input_file_list] to [output_dir]
    Implements single process and multi process (multiprocessing.Pool object must be provided as argument to run_batch
    to run in multi process mode)
    Max number of output files is calculated from length of read_list divided by batch_size
    Every worker receives a single input and single output file (and full set of read_ids to extract)
    If input file is not exhausted before output file has reached batch_size limit, it is given to next worker.
    """
    def __init__(
        self,
        input_file_list: List[Path],
        output_dir: Path,
        read_set: Set[str],
        progressbar: ProgressBar,
        logger: Logger,
        filename_base: str="batch",
        batch_size: int=READS_PER_FILE,
        target_compression: Optional[str]=None,
    ):
        self.input_f5s = input_file_list.copy()  # this list will be modified to track progress
        self.read_set = read_set
        self.target_compression = target_compression
        self.batch_size = batch_size
        self.filename_base = filename_base
        self.output_dir = output_dir
        self.pbar = progressbar
        self.logger = logger

        self.filename_mapping_file = self.output_dir / "filename_mapping.txt"
        if self.filename_mapping_file.exists():
            self.logger.info("overwriting filename mapping file {}".format(self.filename_mapping_file))
            self.filename_mapping_file.unlink()

        self.out_files = {}  # dict where key=filename value=read_set
        self.available_out_files = []
        self.populate_out_files()
        self.tasks = []
        self.pool = None

    def populate_out_files(self) -> None:
        """
        Calculate number of output files based on batch size.
        Delete existing files and initialise dict that keeps track of reads added to each file
        :return:
        """
        num_outputs = int(ceil(len(self.read_set) / float(self.batch_size)))
        for i in range(num_outputs):
            filename = self.filename_base + str(i) + ".fast5"
            output_file_name = self.output_dir / filename

            if output_file_name.exists():
                self.logger.info("overwriting multiread file {}".format(output_file_name))
                output_file_name.unlink()

            self.out_files[output_file_name] = set()
        # reversing so that first item to be popped is lower idx
        self.available_out_files = sorted(self.out_files.keys(), reverse=True)

    def run_batch(self, pool: multiprocessing.Pool=None) -> None:
        """
        Choose sync or async (if multiprocessing.Pool is provided) running mode, launch tasks.
        :param pool:
        :return:
        """
        if pool is None:
            self._launch_sync_tasks()
        else:
            self.pool = pool
            self._launch_async_tasks()

    def _launch_sync_tasks(self) -> None:
        """
        Run tasks sequentially
        :return:
        """
        for args_tuple in self._args_generator():
            reads, out_file, in_file = extract_selected_reads(*args_tuple)
            self._update_file_lists(reads=reads, out_file=out_file, in_file=in_file)

    def _launch_async_tasks(self) -> None:
        """
        Launch an async task for every input-output pair
        self.tasks is just for keeping track of number of tasks still running
        :return:
        """
        for args_tuple in self._args_generator():
            self.pool.apply_async(func=extract_selected_reads, args=args_tuple,
                                  callback=self._callback, error_callback=self._error_callback)

            self.tasks.append(0)

    def _callback(self, result) -> None:
        """
        Once a thread finishes, decrement self.tasks, update available files and reads and trigger scan for new tasks
        :param result: tuple
        :return:
        """
        self._update_file_lists(*result)
        self._launch_async_tasks()
        self.tasks.pop()

    def _error_callback(self, result) -> None:
        self.logger.error(result.original_exception)
        self._update_file_lists(set(), result.output_file, None)
        self._launch_async_tasks()
        self.tasks.pop()

    def _update_file_lists(self, reads, out_file, in_file) -> None:
        """
        Update read sets and files available for processing
        Update progressbar
        :param reads:
        :param out_file:
        :param in_file:
        :return:
        """
        in_file_exhausted = 1
        if in_file is not None:
            # in_file was not exhausted
            in_file_exhausted = 0
            self.input_f5s.append(in_file)

        self.out_files[out_file].update(reads)
        self.read_set.difference_update(reads)

        if len(self.out_files[out_file]) < self.batch_size:
            # out_file has not reached batch limit
            self.available_out_files.append(out_file)

        # print filename - read table
        with open(str(self.filename_mapping_file), 'a') as output_table:
            for read in reads:
                output_table.write("{}\t{}\n".format(read, out_file.name))

        # increment progressbar by number of reads found and by number of files processed
        self.pbar.update(self.pbar.currval + len(reads) + in_file_exhausted)

    def _args_generator(self) -> Iterator[Tuple[Path, Path, Set[str], int, Optional[str]]]:
        """
        If there are possible pairs of input and output files, yield tuples that are suitable inputs to
         extract_selected_reads
        :return: a generator of tuples(in_file, out_file, read_set, count, compression)
        """
        while self.available_out_files and self.input_f5s:
            out_file = self.available_out_files.pop()
            in_file = self.input_f5s.pop()
            count = self.batch_size - len(self.out_files[out_file])
            yield in_file, out_file, self.read_set, count, self.target_compression


def extract_selected_reads(
        input_file: Path,
        output_file: Path,
        read_set: Set[str],
        count: int,
        target_compression: Optional[str]=None,
) -> Tuple[Set[str], Path, Optional[Path]]:
    """
    Take reads from input file if read_id id is in read_set
    Write to output file, at most count times
    return tuple (found_reads, output_file, input_file)
    If input file was exhausted, the third item in return is None
    :param input_file: Path to input Fast5 file
    :param output_file: Path to output Fast5 file (will be appended to if already exists)
    :param read_set: set of read_ids to extract
    :param count: int max number of reads to be added to output_file
    :param target_compression: str type of compression for output fast5 file
    :return: tuple of: found read set, path to output file, path to input file (if not exhausted)
    """
    try:
        found_reads = set()
        with MultiFast5File(str(output_file), 'a') as output_f5:
            reads_present = set(output_f5.get_read_ids())
            for read_id, read in read_generator(input_file, read_set):
                found_reads.add(read_id)

                if read_id in reads_present:
                    continue

                output_f5.add_existing_read(read, target_compression=target_compression)
                reads_present.add(read_id)

                if len(found_reads) >= count:
                    return found_reads, output_file, input_file
    except Exception as e:
        exception = type(e)("Error processing file {}: {}".format(input_file,
                                                                  e.args))
        raise ExtractionException(exception, output_file)

    return found_reads, output_file, None


class ExtractionException(Exception):
    def __init__(self, original_exception, output_file):
        self.original_exception = original_exception
        self.output_file = output_file


def read_generator(input_file: Path, read_set: Set[str]) -> Iterator[Tuple[str, Fast5Read]]:
    """
    Open input_file as Fast5, yield tuples (read_id, fast5_read) for every read_id that is present in read_set
    :param input_file: Path to input Fast5File
    :param read_set: set of read_ids to look for
    :return: tuples of (read_id, read object)
    """
    with get_fast5_file(str(input_file)) as input_f5:
        read_ids = input_f5.get_read_ids()
        for read_id in read_set.intersection(read_ids):
            read = input_f5.get_read(read_id)
            yield read_id, read



try:
    from ConfigParser import ConfigParser
except:  # python3
    from configparser import ConfigParser

import os
from shutil import copyfile
from tempfile import NamedTemporaryFile

from ont_fast5_api import CURRENT_FAST5_VERSION
from ont_fast5_api.fast5_info import Fast5Info, _clean
from ont_fast5_api.fast5_file import Fast5File
from test.helpers import TestFast5ApiHelper, test_data


class TestFast5File(TestFast5ApiHelper):

    def test_001_get_latest_analysis(self):
        test_file = os.path.join(test_data, 'basecall_2d_file_v1.0.fast5')
        with Fast5File(test_file, mode='r') as fh:
            group_name = fh.get_latest_analysis('Basecall_2D')
            self.assertEqual('Basecall_2D_000', group_name)
            # Test a non-existent group.
            group_name = fh.get_latest_analysis('Garbage_5D')
            self.assertEqual(None, group_name)

    def test_002_read_summary_data(self):
        test_file = os.path.join(test_data, 'telemetry_test.fast5')
        summary = Fast5File.read_summary_data(test_file, 'segmentation')
        expected = {'filename': 'telemetry_test.fast5',
                    'channel_id': {u'channel_number': 129,
                                   u'range': 10000.0,
                                   u'sampling_rate': 5000,
                                   u'digitisation': 10000,
                                   u'offset': 0.0},
                    'reads': [{'duration': 755.79559999999947,
                               'start_time': 4034.6948000000002,
                               'read_id': 'telemetry_test.fast5',
                               'start_mux': 1,
                               'read_number': 199}],
                    'tracking_id': {u'device_id': '445444'},
                    'data': {u'split_hairpin': {u'median_sd_comp': 1.4719812720343015,
                                                u'range_comp': 3.965029408419298,
                                                u'median_level_temp': 88.66729546440973,
                                                u'duration_temp': 327.82499999999936,
                                                u'num_temp': 10773,
                                                u'num_events': 24091,
                                                u'median_sd_temp': 1.328457722537222,
                                                u'range_temp': 4.01780031383548,
                                                u'median_level_comp': 89.8680971725336,
                                                u'split_index': 10903,
                                                u'duration_comp': 422.3665999999994,
                                                u'num_comp': 13158},
                             u'empty': {}},
                    'software': {u'time_stamp': '2014-Jun-04 16:28:31',
                                 u'version': '0.5.4',
                                 'component': u'Validation'}}
        self.assertEqual(expected, summary)

    def test_002_add_analysis_group(self):
        fname = self.generate_temp_filename()
        with Fast5File(fname, mode='w') as fast5:
            att = {'foo': 1, 'bar': 2}
            fast5.add_analysis('test', 'Test_000', att)
            att_in = fast5.get_analysis_attributes('Test_000')
            att['component'] = 'test'
            self.assertEqual(att, att_in)
            att2 = {'Bob': 'your uncle'}
            fast5.add_analysis_attributes('Test_000', att2)
            att_in = fast5.get_analysis_attributes('Test_000')
            att.update(att2)
            self.assertEqual(att, att_in)

    def test_003_add_analysis_subgroup(self):
        fname = self.generate_temp_filename()
        with Fast5File(fname, mode='w') as fast5:
            fast5.add_analysis('test', 'Test_000', attrs={})
            fast5.add_analysis_subgroup('Test_000', 'Sub1', attrs={'foo': 'bar', 'monkey': 1})
            att_in = fast5.get_analysis_attributes('Test_000/Sub1')
            self.assertEqual({'foo': 'bar', 'monkey': 1}, att_in)

    def test_010_v0_6_single(self):
        # Check that it is recognized properly.
        fname = os.path.join(test_data, 'read_file_v0.6_single.fast5')
        result = Fast5Info(fname)
        self.assertEqual(0.6, result.version)
        # Copy file and Update to current format.
        new_file = os.path.join(self.save_path, 'single_read_v0.6_test.fast5')
        copyfile(fname, new_file)
        Fast5File.update_legacy_file(new_file)
        result = Fast5Info(new_file)
        self.assertEqual(CURRENT_FAST5_VERSION, result.version)
        self.assertEqual(1, len(result.read_info))
        self.assertEqual(5804, result.read_info[0].read_number)
        # Load the event data.
        with Fast5File(new_file, mode='r') as fh:
            analist = fh.list_analyses('event_detection')
            self.assertEqual(1, len(analist))
            group = '{}/Reads/Read_5804'.format(analist[0][1])
            data = fh.get_analysis_dataset(group, 'Events')
            self.assertEqual(8433, data.size)
            self.assertEqual(set(('mean', 'stdv', 'start', 'length')), set(data.dtype.names))
            read_info = fh.status.read_info[0]
            self.assertEqual(8433, read_info.event_data_count)
            channel_info = fh.get_channel_info()
            self.assertEqual(102, channel_info['channel_number'])

    def test_011_v0_6_raw(self):
        # Check that it is recognized properly.
        fname = os.path.join(test_data, 'read_file_v0.6_raw.fast5')
        result = Fast5Info(fname)
        self.assertEqual(0.6, result.version)
        # Copy file and Update to current format.
        new_file = self.generate_temp_filename()
        copyfile(fname, new_file)
        Fast5File.update_legacy_file(new_file)
        result = Fast5Info(new_file)
        self.assertEqual(CURRENT_FAST5_VERSION, result.version)
        self.assertEqual(1, len(result.read_info))
        self.assertEqual(627, result.read_info[0].read_number)
        # Load the event data.
        with Fast5File(new_file, mode='r') as fh:
            analist = fh.list_analyses('event_detection')
            self.assertEqual(1, len(analist))
            group = '{}/Reads/Read_627'.format(analist[0][1])
            data = fh.get_analysis_dataset(group, 'Events')
            self.assertEqual(2337, data.size)
            self.assertEqual(set(('mean', 'stdv', 'start', 'length')), set(data.dtype.names))
            read_info = fh.status.read_info[0]
            self.assertEqual(2337, read_info.event_data_count)
            channel_info = fh.get_channel_info()
            self.assertEqual(118, channel_info['channel_number'])
            raw = fh.get_raw_data(read_number=627)
            self.assertEqual(46037, raw.size)
            self.assertEqual(46037, read_info.duration)

    def test_012_v1_0_single(self):
        # Check that it is recognized properly.
        fname = os.path.join(test_data, 'read_file_v1.0_single.fast5')
        result = Fast5Info(fname)
        self.assertEqual(1.0, result.version)
        # Copy file and Update to current format.
        new_file = self.generate_temp_filename()
        copyfile(fname, new_file)
        Fast5File.update_legacy_file(new_file)
        result = Fast5Info(new_file)
        self.assertEqual(CURRENT_FAST5_VERSION, result.version)
        self.assertEqual(1, len(result.read_info))
        self.assertEqual(59, result.read_info[0].read_number)
        # Load the event data.
        with Fast5File(new_file, mode='r') as fh:
            analist = fh.list_analyses('event_detection')
            self.assertEqual(1, len(analist))
            group = '{}/Reads/Read_59'.format(analist[0][1])
            data = fh.get_analysis_dataset(group, 'Events')
            self.assertEqual(7875, data.size)
            self.assertEqual(set(('mean', 'stdv', 'start', 'length')), set(data.dtype.names))
            read_info = fh.status.read_info[0]
            self.assertEqual(7875, read_info.event_data_count)
            channel_info = fh.get_channel_info()
            self.assertEqual(1, channel_info['channel_number'])

    def test_fast5_add_and_get_chain(self):
        fname = self.generate_temp_filename()
        group_name1 = 'First_000'
        component1 = 'first'
        component1_path = 'Analyses/{}'.format(group_name1)
        group_name2 = 'Second_000'
        component2 = 'second'

        # Add fake group
        with Fast5File(fname=fname, mode='w') as fast5:
            fast5.add_analysis(component1, group_name1, attrs={})
            fast5.add_analysis(component2, group_name2, attrs={})

            # Check group was added successfully
            target_list_of_analyses = [(component1, group_name1),
                                       (component2, group_name2)]
            self.assertEqual(fast5.list_analyses(), target_list_of_analyses)

            # Check fake group has chain including itself
            target_chain = [(component2, group_name2)]
            self.assertEqual(fast5.get_chain(group_name2), target_chain)

            # Add component chain
            fake_component_map = {component1: group_name1}
            fast5.add_chain(group_name=group_name2,
                            component_map=fake_component_map)

            # Check attributes are as expected
            attr = {'component': component2, component1: component1_path}
            self.assertEqual(fast5.get_analysis_attributes(group_name2), attr)
            # Check chain is as expected
            chain = [(component2, group_name2), (component1, group_name1)]
            self.assertEqual(fast5.get_chain(group_name2), chain)

    def test_fast5_add_group(self):
        fname = self.generate_temp_filename()
        group_name = 'First_000'
        component = 'first'
        component_path = 'Analyses/{}'.format(group_name)

        with Fast5File(fname, mode='w') as fast5:
            # No analyses in file
            self.assertEqual(fast5.list_analyses(), [])
            self.assertEqual(fast5.get_latest_analysis(component), None)
            # Add group to file & check it's there
            fast5._add_group(component_path, {'component': component})
            self.assertEqual(fast5.list_analyses(), [(component, group_name)])
            # Check latest analyses
            target = {group_name: {'component': component}}
            self.assertEqual(fast5._parse_attribute_tree('Analyses'), target)
            # Get latest analyses
            self.assertEqual(fast5.get_latest_analysis('First'), group_name)

    def test_fast5_add_analysis_dataset(self):
        fname = self.generate_temp_filename()
        group_name = 'First_000'
        component = 'first'

        with Fast5File(fname, mode='w') as fast5:
            self.assertFalse(fast5.list_analyses())
            with self.assertRaises(KeyError):
                fast5.add_analysis_dataset(group_name=group_name,
                                           dataset_name='Example',
                                           data='hello',
                                           attrs=None)

            fast5.add_analysis(component=component,
                               group_name=group_name,
                               attrs={})
            fast5.add_analysis_dataset(group_name=group_name,
                                       dataset_name='Example',
                                       data='hello',
                                       attrs=None)
            answer = fast5.get_analysis_dataset(group_name=group_name,
                                                dataset_name='Example')
            self.assertEqual(answer, 'hello')

    def test_fast5_set_analysis_config(self):
        fname = self.generate_temp_filename()
        group_name = 'First_000'
        component = 'first'
        with Fast5File(fname, mode='w') as fast5:
            self.assertFalse(fast5.list_analyses())
            with NamedTemporaryFile(dir=self.save_path,
                                    delete=False, mode='w+t') as f:
                config_str = "[section]\nkey=value\nkey1=value1\n"
                f.write(config_str)
                config_path = f.name

            config = ConfigParser()
            config.read(config_path)
            self.assertTrue(config)
            self.assertEqual(config.get('section', 'key'), 'value')
            with self.assertRaises(KeyError):
                fast5.set_analysis_config(group_name, config)

            fast5.add_analysis(component, group_name, {})
            fast5.set_analysis_config(group_name, config)
            get_config = fast5.get_analysis_config(group_name)
            self.assertEqual(get_config['section']['key'], 'value')

        os.remove(fname)
        with Fast5File(fname, mode='w') as fast5:
            self.assertFalse(fast5.list_analyses())
            config = {'section': {'key': 'value', 'key1': 'value1'}}
            with self.assertRaises(KeyError):
                fast5.set_analysis_config(group_name, config)
            fast5.add_analysis(component, group_name, {})
            fast5.set_analysis_config(group_name, config)
            get_config = fast5.get_analysis_config(group_name)
            self.assertEqual(get_config['section']['key'], 'value')



# Current workflows add the component name to the hdf group as
# an attribute. For older files, the following dict allows us
# to work out the component name from the group name.
LEGACY_COMPONENT_NAMES = {'Alignment': 'alignment',
                          'Basecall_1D': 'basecall_1d',
                          'OnlineBasecall': 'basecall_1d',
                          'Basecall_2D': 'basecall_2d',
                          'Calibration_Strand': 'calibration_strand',
                          'EventDetection': 'event_detection',
                          'Segmentation': 'segmentation',
                          'Hairpin_Split': 'segmentation',
                          'Segment_Linear': 'segmentation',
                          'Validation': 'segmentation',
                          'AlignToRef': 'align_to_ref',
                          'Barcoding': 'barcoding',
                          'Classification': 'classification',
                          'Evaluation': 'evaluation',
                          'Multiple_Alignment': 'multiple_alignment',
                          'Squiggle_Map': 'squiggle_map',
                          'Sam_Segmentor': 'sam_segmentor',
                          'arma': 'arma',
                          'Basic_component': 'basic_component',
                          'RawGenomeCorrected':'raw_genome_corrected' # default name from tombo resquiggle command
                          }

supported_modes = ('r', 'r+', 'w', 'w-', 'x', 'a')
mode_docstring = """Supported file modes:
    r        Readonly, file must exist (default)
    r+       Read/write, file must exist
    w        Create file, truncate if exists
    w- or x  Create file, fail if exists
    a        Read/write if exists, create otherwise"""  # Taken from h5py

HARDLINK_GROUPS = ("context_tags", "tracking_id")
OPTIONAL_READ_GROUPS = {'Analyses'}



import os
import numpy
from unittest.mock import patch
from pathlib import Path

from ont_fast5_api.compression_settings import VBZ
from ont_fast5_api.conversion_tools.fast5_subset import Fast5Filter
from ont_fast5_api.conversion_tools.conversion_utils import Fast5FilterWorker, extract_selected_reads, read_generator
from ont_fast5_api.multi_fast5 import MultiFast5File
from ont_fast5_api.fast5_file import Fast5File
from test.helpers import TestFast5ApiHelper, test_data


class TestFast5Subset(TestFast5ApiHelper):
    input_multif5_path = Path(test_data) / "multi_read" / "batch_0.fast5"
    read_set = {"fe85b517-62ee-4a33-8767-41cab5d5ab39", "fe9374ee-b86a-4ca4-81dc-ac06e3297728"}

    def test_read_generator(self):
        count = 0
        for read_id, read in read_generator(input_file=self.input_multif5_path, read_set=self.read_set):
            self.assertIn(read_id, self.read_set)
            count += 1

        self.assertEqual(len(self.read_set), count)

    def _create_read_list_file(self, read_ids):
        output_path = os.path.join(self.save_path, 'read_list.txt')
        with open(output_path, 'w') as fh:
            for read_id in read_ids:
                fh.write(read_id + "\n")
        return output_path

    @patch('ont_fast5_api.conversion_tools.fast5_subset.logging')
    @patch('ont_fast5_api.conversion_tools.fast5_subset.get_progress_bar')
    def test_subset_from_single(self, mock_log, mock_pbar):
        input_path = os.path.join(test_data, "single_reads")
        read_list = self._create_read_list_file(self.read_set)
        f5_filter = Fast5Filter(input_folder=input_path,
                                output_folder=self.save_path,
                                read_list_file=read_list)
        f5_filter.run_batch()

        count = 0
        with MultiFast5File(os.path.join(self.save_path, 'batch0.fast5'), 'r') as output_f5:
            for input_file in os.listdir(input_path):
                with Fast5File(os.path.join(input_path, input_file), 'r') as input_f5:
                    read_id = input_f5.get_read_id()
                    if read_id in self.read_set:
                        read_in = input_f5.get_read(read_id)
                        read_out = output_f5.get_read(read_id)
                        self.assertTrue(numpy.array_equal(read_in.get_raw_data(), read_out.get_raw_data()))
                        count += 1
        self.assertEqual(len(self.read_set), count)

    @patch('ont_fast5_api.conversion_tools.fast5_subset.logging')
    @patch('ont_fast5_api.conversion_tools.fast5_subset.get_progress_bar')
    def test_subset_from_multi(self, mock_log, mock_pbar):
        read_list = self._create_read_list_file(self.read_set)
        f5_filter = Fast5Filter(input_folder=os.path.dirname(self.input_multif5_path),
                                output_folder=self.save_path,
                                read_list_file=read_list)
        f5_filter.run_batch()
        with MultiFast5File(self.input_multif5_path, 'r') as input_f5, \
                MultiFast5File(os.path.join(self.save_path, 'batch0.fast5'), 'r') as output_f5:
            self.assertEqual(len(self.read_set), len(output_f5.get_read_ids()))
            for read_id in self.read_set:
                read_in = input_f5.get_read(read_id)
                read_out = output_f5.get_read(read_id)
                self.assertTrue(numpy.array_equal(read_in.get_raw_data(), read_out.get_raw_data()))

    def test_extract_selected_reads(self):
        # three test for count below, equaling and above number of read in input file
        for count in (1, 2, 3):
            temp_file_name = self.generate_temp_filename()
            found_reads, output_file, input_file = extract_selected_reads(input_file=self.input_multif5_path,
                                                                          output_file=temp_file_name,
                                                                          count=count, read_set=self.read_set)
            if count < len(self.read_set):
                self.assertTrue(found_reads.issubset(self.read_set))
                self.assertEqual(input_file, self.input_multif5_path)
            elif count == len(self.read_set):
                self.assertEqual(found_reads, self.read_set)
                self.assertEqual(input_file, self.input_multif5_path)
            elif count >= len(self.read_set):
                self.assertEqual(found_reads, self.read_set)
                self.assertIsNone(input_file)

            self.assertEqual(output_file, temp_file_name)
            # verify that resulting output file is a legal MultiFast5 with desired reads in it
            with MultiFast5File(output_file) as multi_file:
                readlist = multi_file.get_read_ids()
                self.assertTrue(set(readlist).issubset(self.read_set))

    @patch('ont_fast5_api.conversion_tools.conversion_utils.ProgressBar')
    @patch('ont_fast5_api.conversion_tools.fast5_subset.logging')
    def test_selector_args_generator(self, mock_pbar, mock_logger):
        single_reads = os.path.join(test_data, "single_reads")
        self.assertTrue(os.path.isdir(single_reads), msg=single_reads)

        input_f5s = list(Path(single_reads).glob('*.fast5'))
        batch_size = 1

        f = Fast5FilterWorker(
            input_file_list=input_f5s,
            output_dir=Path(self.save_path),
            read_set=self.read_set,
            batch_size=batch_size,
            filename_base="batch",
            target_compression=VBZ,
            progressbar=mock_pbar,
            logger=mock_logger
        )

        args_combos = list(f._args_generator())
        # there should be two tuples of arguments
        self.assertEqual(len(args_combos), len(self.read_set) / batch_size)

        num_files_queued = len(f.input_f5s) # should be 0
        self.assertEqual(num_files_queued, (len(input_f5s) - len(args_combos)), msg=f.input_f5s)
        self.assertEqual(len(f.available_out_files), 0)

        # "exhaust" an input file and put output file back on queue
        input_file, output_file, reads, count, compression = args_combos[0]
        f._update_file_lists(reads={}, in_file=None, out_file=output_file)
        self.assertEqual(len(f.input_f5s), num_files_queued)
        self.assertEqual(len(f.available_out_files), 1)
        self.assertEqual(compression, VBZ)

        # this results in another args tuple generated
        new_args_combos = list(f._args_generator())
        self.assertEqual(len(new_args_combos), 1, msg=len(new_args_combos))


"""
Script for binning fast5 reads into separate directories based on column value in summary file
Inteded for demultiplexing reads using barcoding summary file.
"""
from pathlib import Path
from typing import Union, Dict, Set, List
from multiprocessing import Pool
import logging
from csv import reader
from collections import defaultdict
from time import sleep
from math import ceil
from argparse import ArgumentParser

from ont_fast5_api.compression_settings import COMPRESSION_MAP
from ont_fast5_api.conversion_tools.conversion_utils import (
    get_fast5_file_list,
    get_progress_bar,
    Fast5FilterWorker,
    READS_PER_FILE,
    FILENAME_BASE,
    ProgressBar,
)

DEMULTIPLEX_COLUMN = "barcode_arrangement"
READ_ID_COLUMN = "read_id"


class Fast5Demux:
    """
    Bin reads from directory of fast5 files according to demultiplex_column in sequencing_summary path
    :param input_dir: Path to input Fast5 file or directory of Fast5 files
    :param output_dir: Path to output directory
    :param summary_file: Path to TSV summary file
    :param demultiplex_column: str name of column with demultiplex values
    :param read_id_column: str name of column with read ids
    :param filename_base: str prefix for output Fast5 files
    :param batch_size: int maximum number of reads per output file
    :param threads: int maximum number of worker processes
    :param recursive: bool flag to search recursively through input_dir for Fast5 files
    :param follow_symlinks: bool flag to follow symlinks in input_dir
    :param target_compression: str compression type in output Fast5 files
    """

    def __init__(
        self,
        input_dir: Path,
        output_dir: Path,
        summary_file: Path,
        demultiplex_column: str,
        read_id_column: str = READ_ID_COLUMN,
        filename_base: str = FILENAME_BASE,
        batch_size: int = READS_PER_FILE,
        threads: int = 1,
        recursive: bool = False,
        follow_symlinks: bool = True,
        target_compression: Union[str, None] = None,
    ):
        self.input_dir = input_dir
        self.output_dir = output_dir
        self.summary = summary_file
        self.demultiplex_column = demultiplex_column
        self.read_id_column = read_id_column
        self.filename_base = filename_base
        self.batch_size = batch_size
        self.threads = threads
        self.recursive = recursive
        self.follow_symlinks = follow_symlinks
        self.target_compression = target_compression

        self.read_sets: Dict[str, Set[str]] = {}
        self.input_fast5s: List[Path] = []
        self.max_threads: int = 0
        self.workers: List = []
        self.progressbar: Union[ProgressBar, None] = None
        self.logger: logging.Logger = logging.getLogger(self.__class__.__name__)

    def create_output_dirs(self) -> None:
        """
        In output directory create a subdirectory per demux category
        :return:
        """
        self.output_dir.mkdir(parents=True, exist_ok=True)
        for demux in self.read_sets:
            out_dir = self.output_dir / demux
            out_dir.mkdir(exist_ok=True)

    def run_batch(self) -> None:
        """
        Run workers in pool or sequentially
        Starts multiprocessing pool if max_threads allows it
        :return:
        """
        self.workers_setup()

        if self.max_threads > 1:
            with Pool(self.max_threads) as pool:
                for worker in self.workers:
                    worker.run_batch(pool=pool)
                while any(worker.tasks for worker in self.workers):
                    sleep(1)

            pool.join()
            pool.close()
        else:
            for worker in self.workers:
                worker.run_batch(pool=None)

        self.progressbar.finish()

    def workers_setup(self) -> None:
        """
        Parse input summary and input file list to determine amount of work
        Create output directories and initialise workers
        :return:
        """
        self.read_sets = self.parse_summary_demultiplex()
        self.input_fast5s = get_fast5_file_list(
            input_path=self.input_dir,
            recursive=self.recursive,
            follow_symlinks=self.follow_symlinks,
        )
        self.max_threads = self.calculate_max_threads()
        # progressbar length is total numbers of reads to be extracted plus total number of files to be read
        total_progress = sum(len(item) for item in self.read_sets.values()) + (
            len(self.input_fast5s) * len(self.read_sets)
        )
        self.progressbar = get_progress_bar(num_reads=total_progress)
        self.create_output_dirs()
        for demux in sorted(self.read_sets):
            self.workers.append(
                Fast5FilterWorker(
                    input_file_list=self.input_fast5s,
                    output_dir=self.output_dir / demux,
                    read_set=self.read_sets[demux],
                    progressbar=self.progressbar,
                    logger=self.logger,
                    filename_base=self.filename_base,
                    batch_size=self.batch_size,
                    target_compression=self.target_compression,
                )
            )

    def report(self) -> None:
        """
        Log summary of work done
        :return:
        """
        total_reads = 0
        for idx, _ in enumerate(sorted(self.read_sets)):
            worker = self.workers[idx]
            for file, reads in worker.out_files.items():
                total_reads += len(reads)

        self.logger.info("{} reads extracted".format(total_reads))

        # report reads not found
        reads_to_extract = sum(len(item) for item in self.read_sets.values())
        if reads_to_extract > total_reads:
            self.logger.warning(
                "{} reads not found!".format(reads_to_extract - total_reads)
            )

    def calculate_max_threads(self) -> int:
        """
        Calculate max number of workers based on number of output files, input files and threads argument
        :return: int
        """
        max_inputs_per_worker = len(self.input_fast5s)
        total_outputs = 0
        for read_set in self.read_sets.values():
            outputs = int(ceil(len(read_set) / float(self.batch_size)))
            total_outputs += min(outputs, max_inputs_per_worker)

        return min(self.threads, total_outputs)

    def parse_summary_demultiplex(self) -> Dict[str, Set[str]]:
        """
        Open a TSV file and parse read_id and demultiplex columns into dict {demultiplex: read_id_set}
        :return:
        """
        read_sets = defaultdict(set)
        with open(str(self.summary), "r") as fh:
            read_list_tsv = reader(fh, delimiter="\t")
            header = next(read_list_tsv)

            if self.read_id_column in header:
                read_id_col_idx = header.index(self.read_id_column)
            else:
                raise ValueError(
                    "No '{}' read_id column in header: {}".format(
                        self.read_id_column, header
                    )
                )

            if self.demultiplex_column in header:
                demultiplex_col_idx = header.index(self.demultiplex_column)
            else:
                raise ValueError(
                    "No '{}' demultiplex column in header: {}".format(
                        self.demultiplex_column, header
                    )
                )

            for line in read_list_tsv:
                read_id = line[read_id_col_idx]
                demux = line[demultiplex_col_idx]
                read_sets[demux].add(read_id)

        return read_sets


def create_arg_parser():
    parser = ArgumentParser(
        "Tool for binning reads from a multi_read_fast5_file by column value in summary file"
    )
    parser.add_argument(
        "-i",
        "--input",
        required=True,
        type=Path,
        help="Path to Fast5 file or directory of Fast5 files",
    )
    parser.add_argument(
        "-s",
        "--save_path",
        required=True,
        type=Path,
        help="Directory to output MultiRead subset to",
    )
    parser.add_argument(
        "-l",
        "--summary_file",
        required=True,
        type=Path,
        help="TSV file containing read_id column (sequencing_summary.txt file)",
    )
    parser.add_argument(
        "-f",
        "--filename_base",
        default="batch",
        required=False,
        help="Root of output filename, default='{}' -> '{}0.fast5'".format(
            FILENAME_BASE, FILENAME_BASE
        ),
    )
    parser.add_argument(
        "-n",
        "--batch_size",
        type=int,
        default=READS_PER_FILE,
        required=False,
        help="Number of reads per multi-read file (default {})".format(READS_PER_FILE),
    )
    parser.add_argument(
        "-t",
        "--threads",
        type=int,
        default=1,
        required=False,
        help="Maximum number of parallel processes to use (default 1)",
    )
    parser.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        required=False,
        default=False,
        help="Flag to search recursively through input directory for MultiRead fast5 files",
    )
    parser.add_argument(
        "--ignore_symlinks",
        action="store_true",
        help="Ignore symlinks when searching recursively for fast5 files",
    )
    parser.add_argument(
        "-c",
        "--compression",
        required=False,
        default=None,
        choices=list(COMPRESSION_MAP.keys()) + [None],
        help="Target output compression type. If omitted - don't change compression type",
    )
    parser.add_argument(
        "--demultiplex_column",
        type=str,
        default=DEMULTIPLEX_COLUMN,
        required=False,
        help="Name of column for demultiplexing in summary file (default '{}'".format(
            DEMULTIPLEX_COLUMN
        ),
    )
    parser.add_argument(
        "--read_id_column",
        type=str,
        default=READ_ID_COLUMN,
        required=False,
        help="Name of read_id column in summary file (default '{}'".format(
            READ_ID_COLUMN
        ),
    )
    return parser


def main():
    parser = create_arg_parser()
    args = parser.parse_args()
    if args.compression is not None:
        args.compression = COMPRESSION_MAP[args.compression]

    demux = Fast5Demux(
        input_dir=args.input,
        output_dir=args.save_path,
        summary_file=args.summary_file,
        demultiplex_column=args.demultiplex_column,
        read_id_column=args.read_id_column,
        filename_base=args.filename_base,
        batch_size=args.batch_size,
        threads=args.threads,
        recursive=args.recursive,
        follow_symlinks=not args.ignore_symlinks,
        target_compression=args.compression,
    )
    demux.run_batch()
    demux.report()


if __name__ == "__main__":
    main()



from argparse import ArgumentParser

from ont_fast5_api.compression_settings import COMPRESSION_MAP
from ont_fast5_api.conversion_tools.conversion_utils import yield_fast5_files
from ont_fast5_api.fast5_interface import get_fast5_file


def check_read_compression(read):
    """
    Check the compresion type on the raw data of a read
    :param read: Fast5Read object
    :return: AbstractCompression object
    """
    detected_compression = read.raw_compression_filters
    for compression in COMPRESSION_MAP.values():
        if compression.filter_settings == detected_compression:
            return compression
    return detected_compression


def check_compression(input_path, recursive, follow_symlinks, check_all_reads):
    """
    Check the compression type of the raw data in files in a folder
    :param input_path:
    :param recursive:
    :param follow_symlinks:
    :param check_all_reads: bool - check all reads in a file or just the first
    :return: (Compression, read_id, file_path)
    """
    for input_file in yield_fast5_files(input_path, recursive, follow_symlinks):
        with get_fast5_file(input_file, 'r') as f5:
            for read in f5.get_reads():
                compression = check_read_compression(read)
                yield (compression, read.read_id, input_file)
                if not check_all_reads:
                    break


def main():
    parser = ArgumentParser("Tool for checking the compression type of raw data in fast5 files")
    parser.add_argument('-i', '--input_path', required=True,
                        help="Path to Fast5 file or directory of Fast5 files")
    parser.add_argument('--check_all_reads', action='store_true', required=False, default=False,
                        help="Check all reads in a file individually (default: check only the first read)")
    parser.add_argument('-r', '--recursive', action='store_true', required=False, default=False,
                        help="Search recursively through folders for MultiRead fast5 files")
    parser.add_argument('--ignore_symlinks', action='store_true',
                        help="Ignore symlinks when searching recursively for fast5 files")
    parser.add_argument('--file_list', required=False,
                        help="File containing names of files to search in")
    args = parser.parse_args()
    compression_results = check_compression(args.input_path, args.recursive, not args.ignore_symlinks,
                                            args.check_all_reads)
    for result in compression_results:
        print(result)


if __name__ == '__main__':
    main()



from pathlib import Path
from ont_fast5_api.fast5_read import Fast5Read
from test.helpers import test_data, TestFast5ApiHelper

from ont_fast5_api.conversion_tools.conversion_utils import yield_fast5_files, yield_fast5_reads

class TestFast5ConversionUtilities(TestFast5ApiHelper):
    """
    Test the convenience functions yield_fast5_files and yield_fast5_reads
    """

    def setUp(self) -> None:
        super().setUp()

        # Known good read_ids from test_data/multi_read/batch_0.fast5
        self.read_id_set = {'fe849dd3-63bc-4044-8910-14e1686273bb',
                            'fe85b517-62ee-4a33-8767-41cab5d5ab39'}
        self.read_id_list = ['fe849dd3-63bc-4044-8910-14e1686273bb',
                            'fe85b517-62ee-4a33-8767-41cab5d5ab39']
        self.fast5_path = test_data + "/multi_read/batch_0.fast5"


    def test_yield_fast5_files_from_fast5_file(self):
        f5_gen = yield_fast5_files(self.fast5_path, recursive=False)
        f5_path = next(f5_gen)
        self.assertTrue(Path(f5_path).is_file(), "Filepath is not a file")
        self.assertTrue(f5_path.endswith('.fast5'), "Filepath does not end with fast5 extension")
        self.assertTrue(Path(f5_path).absolute() == Path(self.fast5_path).absolute(),
                        "Direct path did not return itself")

    def test_yield_fast5_files_from_dir(self):
        f5_gen = yield_fast5_files(test_data, recursive=False)

        for f5_path in f5_gen:
            self.assertTrue(Path(f5_path).is_file(), "Filepath is not a file")
            self.assertTrue(f5_path.endswith('.fast5'), "Filepath does not end with fast5 extension")

    def test_yield_fast5_reads_from_fast5_file(self):
        f5_read_gen = yield_fast5_reads(self.fast5_path, recursive=False)
        read_id, read_data = next(f5_read_gen)
        self.assertTrue(read_id is not None, "read_id is None")
        self.assertTrue(isinstance(read_data, Fast5Read), "Return is not Fast5Read instance")

    def test_yield_fast5_reads_from_dir(self):
        f5_read_gen = yield_fast5_reads(test_data, recursive=False)
        read_id, read_data = next(f5_read_gen)
        self.assertTrue(read_id is not None, "read_id is None")
        self.assertTrue(isinstance(read_data, Fast5Read), "Return is not Fast5Read instance")

    def test_yield_fast5_reads_with_set(self):
        f5_read_gen = yield_fast5_reads(self.fast5_path,
                                        recursive=False,
                                        read_ids=self.read_id_set)
        f5_reads = list(f5_read_gen)
        self.assertTrue(len(f5_reads) == len(self.read_id_set))

        for read_id, read_data in f5_reads:
            self.assertTrue(read_id in self.read_id_set, "A read_id is not a member of read_ids")
            self.assertTrue(isinstance(read_data, Fast5Read), "Return is not Fast5Read instance")

    def test_yield_fast5_reads_with_list(self):
        f5_read_gen = yield_fast5_reads(self.fast5_path,
                                        recursive=False,
                                        read_ids=self.read_id_set)
        f5_reads = list(f5_read_gen)
        self.assertTrue(len(f5_reads) == len(self.read_id_list))

        for read_id, read_data in f5_reads:
            self.assertTrue(read_id in self.read_id_set, "A read_id is not a member of read_id_list")
            self.assertTrue(isinstance(read_data, Fast5Read), "Return is not Fast5Read instance")

    def test_yield_fast5_reads_set_versus_list_equality(self):
        f5_read_gen_by_id_set = yield_fast5_reads(self.fast5_path,
                                                  recursive=False,
                                                  read_ids=self.read_id_set)

        f5_read_gen_by_id_list = yield_fast5_reads(self.fast5_path,
                                                   recursive=False,
                                                   read_ids=self.read_id_list)

        # Consume the generators into sets
        ids_by_set = set(rid for rid, _ in f5_read_gen_by_id_set)
        ids_by_list = set(rid for rid, _  in f5_read_gen_by_id_list)
        self.assertTrue(ids_by_list == ids_by_set, 'Ids differ when using read_id list versus set')


    def test_yield_fast5_reads_with_empty_set(self):
        f5_read_gen = yield_fast5_reads(self.fast5_path,
                                        recursive=False,
                                        read_ids=set([]))

        self.assertTrue(len(list(f5_read_gen)) != 0, "Empty read_ids resulted in zero returned reads")

    def test_yield_fast5_reads_with_garbage_set(self):
        f5_read_gen = yield_fast5_reads(self.fast5_path,
                                        recursive=False,
                                        read_ids={'_g4rbag£_'})
        f5_reads = list(f5_read_gen)
        self.assertTrue(len(f5_reads) == 0, "Garbage read_ids returned non-zero reads")

    def test_yield_fast5_reads_type_error(self):
        with self.assertRaisesRegex(TypeError, 'read_ids'):
            f5_read_gen = yield_fast5_reads(self.fast5_path,
                                            recursive=False,
                                            read_ids=int(1))
            next(f5_read_gen)


import logging
import os
import shutil
from argparse import ArgumentParser, ArgumentError
from multiprocessing.pool import Pool

from ont_fast5_api import __version__
from ont_fast5_api.compression_settings import COMPRESSION_MAP
from ont_fast5_api.conversion_tools.conversion_utils import get_fast5_file_list, get_progress_bar
from ont_fast5_api.fast5_file import Fast5File, EmptyFast5
from ont_fast5_api.fast5_read import copy_attributes
from ont_fast5_api.fast5_interface import is_multi_read
from ont_fast5_api.multi_fast5 import MultiFast5File
from ont_fast5_api.static_data import OPTIONAL_READ_GROUPS


def compress_batch(input_folder, output_folder, target_compression, recursive=True, threads=1, follow_symlinks=True,
                   in_place=False, sanitize=False):
    # We require an absolute input path to we can replicate the data structure relative to it later on
    input_folder = os.path.abspath(input_folder)

    file_list = get_fast5_file_list(input_folder, recursive, follow_symlinks=follow_symlinks)
    if len(file_list) == 0:
        raise ValueError("No input fast5 files found in '{}'. Recursive={}".format(input_folder, recursive))

    # Set up the process pool and the progressbar
    pool = Pool(min(threads, len(file_list)))
    pbar = get_progress_bar(len(file_list))

    def update(result):
        if in_place and result is not None:
            input_path, output_path = result
            shutil.move(output_path, input_path)
        pbar.update(pbar.currval + 1)

    for input_file in file_list:
        input_path = os.path.join(input_folder, input_file)
        if in_place:
            output_path = os.path.join(input_path + ".tmp.compressed")
        else:
            output_path = os.path.join(output_folder, os.path.relpath(input_path, input_folder))

        pool.apply_async(func=compress_file,
                         args=(input_path, output_path, target_compression, sanitize),
                         callback=update)

    # Tear down the process pool and pbar. We can't use contextmanagers since we need to close() then join()
    pool.close()
    pool.join()
    pbar.finish()


def compress_file(input_file, output_file, target_compression, sanitize=False):
    try:
        os.makedirs(os.path.dirname(output_file), exist_ok=True)
        if is_multi_read(input_file):
            with MultiFast5File(input_file, 'r') as input_f5, MultiFast5File(output_file, 'a') as output_f5:
                for read in input_f5.get_reads():
                    output_f5.add_existing_read(read, target_compression, sanitize=sanitize)
        else:
            with Fast5File(input_file, 'r') as input_f5, \
                    EmptyFast5(output_file, 'a') as output_f5:
                compress_single_read(output_f5, input_f5, target_compression, sanitize=sanitize)
    except Exception as e:
        # Error raised in Pool.async will be lost so we explicitly print them.
        logging.exception(e)
        raise
    return (input_file, output_file)


def compress_single_read(output_f5, read_to_copy, target_compression, sanitize=False):
    read_id = read_to_copy.get_read_id()
    raw_dataset_name = read_to_copy.raw_dataset_name
    raw_group_name = read_to_copy.raw_dataset_group_name
    read_name = "read_" + read_id
    # Recreating the status object is painful, but doesn't actually interact with the file so we can just reference it.
    output_f5.status = read_to_copy.status

    if str(target_compression) in read_to_copy.raw_compression_filters:
        # If we have the right compression then no need for doing anything fancy
        output_f5.handle.copy(read_to_copy.handle, read_name)
    else:
        copy_attributes(read_to_copy.handle.attrs, output_f5.handle)
        for subgroup in read_to_copy.handle:
            if subgroup not in raw_dataset_name:
                if sanitize and subgroup in OPTIONAL_READ_GROUPS:
                    # skip optional groups when sanitizing
                    continue
                output_f5.handle.copy(read_to_copy.handle[subgroup], subgroup)
            else:
                raw_attrs = read_to_copy.handle[raw_group_name].attrs
                raw_data = read_to_copy.handle[raw_dataset_name]
                output_f5.add_raw_data(raw_data, raw_attrs, compression=target_compression)


def main():
    parser = ArgumentParser("Tool for changing the compression of Fast5 files")
    parser.add_argument('-i', '--input_path', required=True,
                        help='Folder containing fast5 files')

    output_group = parser.add_mutually_exclusive_group(required=True)
    save_arg = output_group.add_argument('-s', '--save_path', default=None,
                                         help="Folder to output fast5 read files to")
    output_group.add_argument('--in_place', action='store_true',
                              help='Replace the old files with new files in place')

    parser.add_argument('-c', '--compression', required=True, choices=list(COMPRESSION_MAP.keys()),
                        help="Target output compression type")
    parser.add_argument('--sanitize', action='store_true',
                        help="Clean output files of optional groups and datasets (e.g. 'Analyses')")
    parser.add_argument('-t', '--threads', type=int, default=1, required=False,
                        help="Maximum number of threads to use")
    parser.add_argument('--recursive', action='store_true',
                        help="Search recursively through folders for single_read fast5 files")
    parser.add_argument('--ignore_symlinks', action='store_true',
                        help="Ignore symlinks when searching recursively for fast5 files")
    parser.add_argument('-v', '--version', action='version', version=__version__)
    args = parser.parse_args()

    if args.input_path == args.save_path:
        raise ArgumentError(save_arg, "--input_path and --save_path must be different locations, or use --in_place")
    if args.sanitize and args.save_path is None:
        raise ArgumentError(save_arg, "--save_path must be given if using --sanitize")

    compress_batch(input_folder=args.input_path,
                   output_folder=args.save_path,
                   target_compression=COMPRESSION_MAP[args.compression],
                   threads=args.threads,
                   recursive=args.recursive,
                   follow_symlinks=not args.ignore_symlinks,
                   in_place=args.in_place,
                   sanitize=args.sanitize)


if __name__ == '__main__':
    main()



import h5py

from ont_fast5_api import CURRENT_FAST5_VERSION
from ont_fast5_api.fast5_file import Fast5File, Fast5FileTypeError
from ont_fast5_api.fast5_read import AbstractFast5, Fast5Read, copy_attributes
from ont_fast5_api.static_data import HARDLINK_GROUPS, OPTIONAL_READ_GROUPS


class MultiFast5File(AbstractFast5):
    def __init__(self, filename, mode='r', driver=None):
        # See https://docs.h5py.org/en/stable/high/file.html#file-drivers for
        # information on why you might want to use a specific driver.
        self.filename = filename
        self.mode = mode
        self.handle = h5py.File(self.filename, self.mode, driver=driver)
        self._run_id_map = None
        if mode != 'r' and 'file_version' not in self.handle.attrs:
            try:
                self.handle.attrs['file_version'] = str(CURRENT_FAST5_VERSION)
            except IOError as e:
                raise IOError("Could not write 'file_version' in mode '{}' to fast5 file: {}"
                                   "".format(self.filename, self.mode)) from e

    def get_reads(self):
        for group_name in self.handle:
            if group_name.startswith('read_'):
                yield Fast5Read(self, group_name[5:])

    def get_read_ids(self):
        # Return all groups with the 'read_' stripped from the front
        return [group_name[5:] for group_name in self.handle if group_name.startswith('read_')]

    def get_read(self, read_id):
        group_name = "read_" + read_id
        if group_name not in self.handle:
            raise KeyError("Read '{}' not in: {}".format(group_name, self.filename))
        return Fast5Read(self, read_id)

    def create_read(self, read_id, run_id):
        DeprecationWarning("'MultiFast5File.create_read()' will be deprecated. "
                           "Use `MultiFast5File.create_empty_read()` instead")
        return self.create_empty_read(read_id, run_id)

    def create_empty_read(self, read_id, run_id):
        group_name = "read_" + read_id
        if group_name not in self.handle:
            try:
                self.handle.create_group(group_name)
            except ValueError as e:
                raise ValueError("Could not create group '{}' in file: {}".format(group_name, self.filename)) from e
            try:
                for shared_group in HARDLINK_GROUPS:
                    self.handle["{}/{}".format(group_name, shared_group)] = \
                        self.handle["read_{}/{}".format(self.run_id_map[run_id], shared_group)]
            except KeyError:
                # If we can't hardlink to existing groups then continue as normal
                # registering this read as the new source of metadata for this run_id_map
                self.run_id_map[run_id] = read_id
            self.handle[group_name].attrs["run_id"] = run_id
        else:
            raise ValueError("Read '{}' already exists in: {}".format(group_name, self.filename))
        return Fast5Read(self, read_id)


    @property
    def run_id_map(self):
        if self._run_id_map is None:
            self._run_id_map = dict()
            for read in self.get_reads():
                try:
                    self._run_id_map[read.run_id] = read.read_id
                except KeyError:
                    # If we can't find the read.run_id then there is a KeyError
                    # We want to ignore these cases since they can't be linked anyway
                    pass
        return self._run_id_map

    def add_existing_read(self, read_to_add, target_compression=None, sanitize=False):
        if isinstance(read_to_add, Fast5File):
            self._add_read_from_single(read_to_add, target_compression, sanitize=sanitize)
        elif isinstance(read_to_add.parent, MultiFast5File):
            self._add_read_from_multi(read_to_add, target_compression, sanitize=sanitize)
        else:
            raise Fast5FileTypeError("Could not add read to output file from input file type '{}' with path '{}'"
                                     "".format(type(read_to_add.parent), read_to_add.parent.filename))

    def _add_read_from_multi(self, read_to_add, target_compression, sanitize=False):
        read_name = "read_" + read_to_add.read_id
        self.handle.create_group(read_name)
        output_group = self.handle[read_name]
        copy_attributes(read_to_add.handle.attrs, output_group)
        for subgroup in read_to_add.handle:
            if sanitize and subgroup in OPTIONAL_READ_GROUPS:
                # skip optional groups when sanitizing
                continue
            elif subgroup == read_to_add.raw_dataset_group_name \
                    and target_compression is not None \
                    and str(target_compression) not in read_to_add.raw_compression_filters:
                raw_attrs = read_to_add.handle[read_to_add.raw_dataset_group_name].attrs
                raw_data = read_to_add.handle[read_to_add.raw_dataset_name]
                output_read = self.get_read(read_to_add.read_id)
                output_read.add_raw_data(raw_data, raw_attrs, compression=target_compression)
                continue
            elif subgroup in HARDLINK_GROUPS:
                if read_to_add.run_id in self.run_id_map:
                    # There may be a group to link to, but we must check it actually exists!
                    hardlink_source = "read_{}/{}".format(self.run_id_map[read_to_add.run_id], subgroup)
                    if hardlink_source in self.handle:
                        hardlink_dest = "read_{}/{}".format(read_to_add.read_id, subgroup)
                        self.handle[hardlink_dest] = self.handle[hardlink_source]
                        continue
                # If we couldn't hardlink to anything we need to make the group we create available for future reads
                self.run_id_map[read_to_add.run_id] = read_to_add.read_id
            # If we haven't done a special-case copy then we can fall back on the default copy
            output_group.copy(read_to_add.handle[subgroup], subgroup)

    def _add_read_from_single(self, read_to_add, target_compression, sanitize=False):
        read_name = "read_" + read_to_add.read_id
        self.handle.create_group(read_name)
        output_group = self.handle[read_name]
        copy_attributes(read_to_add.handle.attrs, output_group)
        for subgroup in read_to_add.handle:
            if sanitize and subgroup in OPTIONAL_READ_GROUPS:
                # skip optional groups when sanitizing
                continue
            elif subgroup == "UniqueGlobalKey":
                for unique_group in read_to_add.handle["UniqueGlobalKey"]:
                    if unique_group in HARDLINK_GROUPS and read_to_add.run_id in self.run_id_map:
                        hardlink_source = "read_{}/{}".format(self.run_id_map[read_to_add.run_id], unique_group)
                        if hardlink_source in self.handle:
                            hardlink_dest = "read_{}/{}".format(read_to_add.read_id, unique_group)
                            self.handle[hardlink_dest] = self.handle[hardlink_source]
                    else:
                        output_group.copy(read_to_add.handle["UniqueGlobalKey/{}".format(unique_group)],
                                          unique_group)
                self.run_id_map[read_to_add.run_id] = read_to_add.read_id
            elif subgroup == "Raw":
                if target_compression is None or str(target_compression) in read_to_add.raw_compression_filters:
                    output_group.copy(read_to_add.handle[read_to_add.raw_dataset_group_name], "Raw")
                else:
                    raw_attrs = read_to_add.handle[read_to_add.raw_dataset_group_name].attrs
                    raw_data = read_to_add.handle[read_to_add.raw_dataset_name]
                    output_read = self.get_read(read_to_add.read_id)
                    output_read.add_raw_data(raw_data, raw_attrs, compression=target_compression)
            else:
                if not sanitize:
                    output_group.copy(read_to_add.handle[subgroup], subgroup)



from argparse import ArgumentParser
from multiprocessing import Pool
import logging
import os

from ont_fast5_api import __version__
from ont_fast5_api.conversion_tools.conversion_utils import get_fast5_file_list, get_progress_bar
from ont_fast5_api.fast5_file import EmptyFast5, Fast5FileTypeError
from ont_fast5_api.fast5_interface import check_file_type, MULTI_READ
from ont_fast5_api.multi_fast5 import MultiFast5File

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)
exc_info = False


def batch_convert_multi_files_to_single(input_path, output_folder, threads, recursive, follow_symlinks):
    pool = Pool(threads)
    file_list = get_fast5_file_list(input_path, recursive, follow_symlinks=follow_symlinks)
    pbar = get_progress_bar(len(file_list))

    def update(result):
        input_file = result[0]
        with open(os.path.join(output_folder, "filename_mapping.txt"), 'a') as output_table:
            for filename in result[1]:
                output_table.write("{}\t{}\n".format(input_file, filename))
        pbar.update(pbar.currval + 1)

    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    results_array = []
    for batch_num, filename in enumerate(file_list):
        results_array.append(pool.apply_async(convert_multi_to_single,
                                              args=(filename, output_folder,
                                                    str(batch_num)),
                                              callback=update))

    pool.close()
    pool.join()
    pbar.finish()


def convert_multi_to_single(input_file, output_folder, subfolder):
    output_files = ()
    try:
        output_files = try_multi_to_single_conversion(input_file, output_folder, subfolder)
    except Exception as e:
        logger.error("{}\n\tFailed to copy files from: {}"
                     "".format(e, input_file), exc_info=exc_info)
    return input_file, output_files


def try_multi_to_single_conversion(input_file, output_folder, subfolder):
    output_files = []
    with MultiFast5File(input_file, 'r') as multi_f5:
        file_type = check_file_type(multi_f5)
        if file_type != MULTI_READ:
            raise Fast5FileTypeError("Could not convert Multi->Single for file type '{}' with path '{}'"
                                     "".format(file_type, input_file))
        for read in multi_f5.get_reads():
            try:
                output_file = os.path.join(output_folder, subfolder, "{}.fast5".format(read.read_id))
                create_single_f5(output_file, read)
                output_files.append(os.path.basename(output_file))
            except Exception as e:
                logger.error("{}\n\tFailed to copy read '{}' from {}"
                             "".format(str(e), read.read_id, input_file), exc_info=exc_info)
    return output_files


def create_single_f5(output_file, read):
    if not os.path.exists(os.path.dirname(output_file)):
        os.makedirs(os.path.dirname(output_file))
    with EmptyFast5(output_file, 'w') as single_f5:
        for group in read.handle:
            if group == "Raw":
                read_number = read.handle["Raw"].attrs["read_number"]
                single_f5.handle.copy(read.handle[group], "Raw/Reads/Read_{}".format(read_number))
            elif group in ("channel_id", "context_tags", "tracking_id"):
                if "UniqueGlobalKey" not in single_f5.handle:
                    single_f5.handle.create_group("UniqueGlobalKey")
                single_f5.handle.copy(read.handle[group], "UniqueGlobalKey/{}".format(group))
            else:
                single_f5.handle.copy(read.handle[group], group)


def main():
    parser = ArgumentParser("")
    parser.add_argument('-i', '--input_path', required=True,
                        help="MultiRead fast5 file or path to directory of MultiRead files")
    parser.add_argument('-s', '--save_path', required=True,
                        help="Folder to output SingleRead fast5 files to")
    parser.add_argument('--recursive', action='store_true',
                        help="Search recursively through folders for MultiRead fast5 files")
    parser.add_argument('--ignore_symlinks', action='store_true',
                        help="Ignore symlinks when searching recursively for fast5 files")
    parser.add_argument('-t', '--threads', type=int, default=1, required=False,
                        help="Number of threads to use")
    parser.add_argument('-v', '--version', action='version', version=__version__)
    args = parser.parse_args()

    batch_convert_multi_files_to_single(args.input_path, args.save_path, args.threads,
                                        args.recursive, follow_symlinks=not args.ignore_symlinks)


if __name__ == '__main__':
    main()



import h5py
import os
import shutil
from unittest.mock import patch

from ont_fast5_api.compression_settings import VBZ, GZIP
from ont_fast5_api.conversion_tools.check_file_compression import check_read_compression, check_compression
from ont_fast5_api.conversion_tools.compress_fast5 import compress_file, compress_single_read, compress_batch
from ont_fast5_api.conversion_tools.conversion_utils import get_fast5_file_list
from ont_fast5_api.fast5_file import Fast5File, EmptyFast5
from ont_fast5_api.fast5_info import ReadInfo
from ont_fast5_api.fast5_interface import get_fast5_file
from ont_fast5_api.multi_fast5 import MultiFast5File
from ont_fast5_api.static_data import OPTIONAL_READ_GROUPS
from test.helpers import TestFast5ApiHelper, test_data


class TestVbzReadWrite(TestFast5ApiHelper):
    run_id = "123abc"

    def test_write_vbz_directly(self):
        input_data = range(10)
        with h5py.File(os.path.join(self.save_path, 'h5py.fast5'), 'w') as fast5:
            fast5.create_dataset('Signal', data=input_data, **vars(VBZ))
            raw = fast5['Signal']

            self.assertTrue(str(VBZ.compression) in raw._filters)
            self.assertEqual(VBZ.compression_opts, raw._filters[str(VBZ.compression)])
            self.assertEqual(list(input_data), list(raw))

    def test_read_vbz_using_api(self):
        with MultiFast5File(os.path.join(test_data, 'vbz_reads', 'vbz_reads.fast5'), 'r') as fast5:
            read_count = 0
            for read in fast5.get_reads():
                # This input file was created to have 4 reads with 20 samples per read
                read_count += 1
                raw_data = read.get_raw_data()
                self.assertEqual(20, len(raw_data))
            self.assertEqual(4, read_count)

    def test_write_vbz_using_api(self):
        input_data = list(range(5))
        read_id = "0a1b2c3d"
        with MultiFast5File(self.generate_temp_filename(), 'w') as fast5:
            fast5.create_empty_read(read_id, self.run_id)
            read = fast5.get_read(read_id)
            read.add_raw_data(input_data, attrs={}, compression=VBZ)
            raw = read.get_raw_data()
            # First check the data comes back in an appropriate form
            self.assertEqual(input_data, list(raw))
            # Then check the types are as they should be under the hood
            filters = read.raw_compression_filters
            self.assertTrue(str(VBZ.compression) in filters)
            self.assertEqual(VBZ.compression_opts, filters[str(VBZ.compression)])

    def test_write_vbz_using_api_single_read(self):
        input_data = list(range(5))
        read_id = "0a1b2c3d"
        read_number = 0
        with Fast5File(self.generate_temp_filename(), 'w') as fast5:
            fast5.status.read_number_map[read_number] = read_number
            fast5.status.read_info = [ReadInfo(read_number=read_number, read_id=read_id,
                                               start_time=0, duration=len(input_data))]
            fast5.add_raw_data(data=input_data, attrs={}, compression=VBZ)
            raw = fast5.get_raw_data()
            # First check the data comes back in an appropriate form
            self.assertEqual(input_data, list(raw))

            # Then check the types are as they should be under the hood
            filters = fast5.raw_compression_filters
            self.assertTrue(str(VBZ.compression) in filters)
            self.assertEqual(VBZ.compression_opts, filters[str(VBZ.compression)])


class TestVbzConvert(TestFast5ApiHelper):
    run_id = "123abc"

    def assertCompressed(self, data_path, expected_compression, read_count, file_count):
        files = set()
        read_ids = set()
        for compression, read_id, filepath in check_compression(data_path, False, False, check_all_reads=True):
            self.assertEqual(expected_compression, compression)
            read_ids.add(read_id)
            files.add(filepath)
        self.assertEqual(read_count, len(read_ids))
        self.assertEqual(file_count, len(files))

    def assert_end_reason_attr_is_enum(self, _: str, obj: h5py.HLObject):
        """
        Assert that the end_reason attribute is an h5 enumeration

        Parameters matches signature for h5py.Group.visititems(name, object) used in the
        recursive search. name is unused.
        """
        if "end_reason" in obj.attrs:
            # Set state variable that end_reason attribute is seen in the given file
            # This is used to assert that this test doesn't pass due to an absence of a
            # negative result
            self.end_reason_seen = True

            metadata = obj.attrs.get_id("end_reason").dtype.metadata
            self.assertIsNotNone(
                metadata,
                msg="end_reason attribute dtype.metadata is None.  "
                    "This indicates that the enumeration metadata is lost"
            )
            self.assertTrue(
                "enum" in metadata,
                msg="end_reason attribute dtype.metadata does not contain enum"
            )

    def test_add_read_from_multi(self):
        target_compression = VBZ
        with get_fast5_file(os.path.join(test_data, "multi_read", "batch_0.fast5"), "r") as input_f5, \
                MultiFast5File(self.generate_temp_filename(), 'w') as output_f5:
            read_id = input_f5.get_read_ids()[0]
            input_read = input_f5.get_read(read_id)

            # Input read should be uncompressed on the way in:
            self.assertEqual(check_read_compression(input_read), GZIP)

            output_f5.add_existing_read(input_read, target_compression)

            output_read = output_f5.get_read(read_id)
            self.assertEqual(check_read_compression(output_read), VBZ)

    def test_compress_read_from_single(self):
        with get_fast5_file(os.path.join(test_data, "single_reads", "read0.fast5"), "r") as input_f5, \
                EmptyFast5(self.generate_temp_filename(), 'w') as output_f5:
            read_id = input_f5.get_read_ids()[0]
            input_read = input_f5.get_read(read_id)

            # Input read should be uncompressed on the way in:
            self.assertEqual(check_read_compression(input_read), GZIP)

            compress_single_read(output_f5, input_read, target_compression=VBZ)

            output_read = output_f5.get_read(read_id)
            self.assertEqual(check_read_compression(output_read), VBZ)

    @patch('ont_fast5_api.conversion_tools.compress_fast5.get_progress_bar')
    def test_conversion_script_multi(self, mock_pbar):
        input_folder = os.path.join(test_data, 'multi_read')
        compress_batch(input_folder=input_folder, output_folder=self.save_path, target_compression=VBZ)
        self.assertCompressed(self.save_path, VBZ, read_count=4, file_count=1)

    @patch('ont_fast5_api.conversion_tools.compress_fast5.get_progress_bar')
    def test_conversion_script_single(self, mock_pbar):
        input_folder = os.path.join(test_data, 'single_reads')
        compress_batch(input_folder=input_folder, output_folder=self.save_path, target_compression=VBZ)
        self.assertCompressed(self.save_path, VBZ, read_count=4, file_count=4)

    def assert_all_files_retain_end_reason_enumeration_metadata(self):
        """
        Assertion that all output files have not lost the end_reason enumeration metadata
        """

        for f5_file in get_fast5_file_list(self.save_path, recursive=True):

            # Require that the end_reason attribute is seen in the file under test.
            # This is used to assert that this test doesn't pass due to an absence of a
            # negative result
            self.end_reason_seen = False

            with h5py.File(f5_file, "r") as fh:
                # Check the top level item
                self.assert_end_reason_attr_is_enum("", fh)

                # Recursively search all items in the file
                fh.visititems(self.assert_end_reason_attr_is_enum)

            self.assertTrue(
                self.end_reason_seen,
                msg="No end_reason attributes seen in the fast5 file. Something went wrong."
            )

    @patch('ont_fast5_api.conversion_tools.compress_fast5.get_progress_bar')
    def test_conversion_script_multi_retains_end_reason_enumeration(self, mock_pbar):
        """
        Test that given a collection of multi read fast5 files that undergo compression
        that the end_reason attribute metadata which indicates that it is an enumeration,
        is not lost
        """
        input_folder = os.path.join(test_data, 'multi_read')
        compress_batch(input_folder=input_folder, output_folder=self.save_path,
                       target_compression=VBZ)

        self.assert_all_files_retain_end_reason_enumeration_metadata()

    @patch('ont_fast5_api.conversion_tools.compress_fast5.get_progress_bar')
    def test_conversion_script_single_retains_end_reason_enumeration(self, mock_pbar):
        """
        Test that given a collection of single read fast5 files that undergo compression
        that the end_reason attribute metadata which indicates that it is an enumeration,
        is not lost
        """

        input_folder = os.path.join(test_data, 'single_reads')
        compress_batch(input_folder=input_folder, output_folder=self.save_path,
                       target_compression=VBZ)

        self.assert_all_files_retain_end_reason_enumeration_metadata()

    @patch('ont_fast5_api.conversion_tools.compress_fast5.get_progress_bar')
    def test_compress_in_place(self, mock_pbar):
        for input_file in os.listdir(os.path.join(test_data, 'single_reads')):
            # We copy file by file as copytree won't work to an existing directory
            shutil.copy(os.path.join(test_data, 'single_reads', input_file), self.save_path)

        self.assertCompressed(self.save_path, GZIP, read_count=4, file_count=4)
        in_files = set(os.listdir(self.save_path))
        compress_batch(self.save_path, output_folder=None, target_compression=VBZ, in_place=True)
        self.assertCompressed(self.save_path, VBZ, read_count=4, file_count=4)
        self.assertEqual(in_files, set(os.listdir(self.save_path)))


class TestSanitise(TestFast5ApiHelper):
        
    @staticmethod
    def list_groups(fname, single_multi='multi'):
        split_index = {
            'multi': 1, 'single': 0}
        all_groups = list()
        filtered_groups = list()
        def _add_group(name):
            all_groups.append(name)
            try:
                subgroup = name.split('/')[split_index[single_multi]]
            except IndexError:
                # top level
                filtered_groups.append(name)
            else:
                if not subgroup in OPTIONAL_READ_GROUPS:
                    filtered_groups.append(name)
        with h5py.File(fname, 'r') as fh:
            fh.visit(_add_group)
        return all_groups, filtered_groups

    def _test(self, input_file, output_file, single_or_multi):
        orig_all_groups, orig_filtered_groups = self.list_groups(input_file, single_or_multi)
        new_all_groups, new_filtered_groups = self.list_groups(output_file, single_or_multi)
        
        self.assertNotEqual(orig_all_groups, orig_filtered_groups)
        self.assertEqual(orig_filtered_groups, new_filtered_groups)
        self.assertEqual(new_all_groups, new_filtered_groups)

    def test_multi_to_multi(self):
        input_file = os.path.join(test_data, "multi_read_analyses", "batch_0.fast5")
        output_file = self.generate_temp_filename()
        compress_file(input_file, output_file, VBZ, sanitize=True)
        self._test(input_file, output_file, 'multi')

    def test_single_to_multi(self):
        input_file = os.path.join(test_data, "single_read_analyses", "read.fast5")
        output_file = self.generate_temp_filename()
        with Fast5File(input_file, 'r') as input_f5, \
                EmptyFast5(output_file, 'a') as output_f5:
            compress_single_read(output_f5, input_f5, VBZ, sanitize=True)
        self._test(input_file, output_file, 'single')




import unittest
from unittest.mock import patch
import os

from ont_fast5_api.compression_settings import register_plugin

class TestCheckCompression(unittest.TestCase):

    def test_register_plugin_no_prepend(self):
        # GIVEN a version of h5py that has the h5pl module
        with patch('h5py.h5pl', create=True) as mock:

            # WHEN h5py.h5pl doesn't have the prepend method
            del mock.prepend
            # and the hdf5 plugin path variable hasn't been set
            if 'HDF5_PLUGIN_PATH' in os.environ:
                del os.environ['HDF5_PLUGIN_PATH']
            self.assertTrue('HDF5_PLUGIN_PATH' not in os.environ)

            # THEN when we try and register the vbz plugin we fall back to setting
            # HDF5_PLUGIN_PATH (because we can't use prepend)
            plugin_path = register_plugin()
            self.assertTrue(os.environ['HDF5_PLUGIN_PATH'] == plugin_path)
        
            



import abc
from abc import abstractmethod

from ont_fast5_api.fast5_file import Fast5File, Fast5FileTypeError
from ont_fast5_api.fast5_read import Fast5Read


class BaseTool(object):
    @property
    def group_id(self):
        raise NotImplementedError("BaseTool does not have a group_id")

    @property
    def analysis_id(self):
        raise NotImplementedError("BaseTool does not have a analysis_id")

    def __init__(self, source, mode='r', group_name=None, meta=None, config=None):
        """ Create a new analysis_tools object.

        :param source: Either an open Fast5File object, or a filename
            of a fast5 file.
        :param mode: The open mode (r or r+). Only if a filename is used
            for the source argument.
        :param group_name: The specific analysis instance you are interested in.
        :param meta: Metadata for a new analysis.
        :param config: Configuration data for a new analysis.

        To create a new analysis group, provide a group name that
        does not already exist, and an optional dictionary with the metadata.
        The following fields are recommended, as a minimum:

            * name - The name of the software used.
            * time_stamp - The time at which the analysis was performed.

        If the group name already exists, the "meta" parameter is ignored. If
        the specified group has a "component" attribute, and its value does not
        match self.analysis_id, an exception will be thrown.
        """
        if isinstance(source, Fast5Read):
            self.filename = source.filename  # Useful for debugging purposes
            self.handle = source
            self.close_handle_when_done = False
        elif isinstance(source, str):
            self.filename = source  # Useful for debugging purposes
            try:
                self.handle = Fast5File(source, mode)
            except Fast5FileTypeError :
                raise NotImplementedError("AnalysisTools do not support accessing MultiReadFast5 files by filepath")
            self.close_handle_when_done = True
        else:
            raise KeyError('Unrecognized type for argument "source": {}'.format(source))
        if group_name is None:
            group_name = self.handle.get_latest_analysis(self.group_id)
            if group_name is None:
                raise KeyError('No group: {} found in file: {}'.format(group_name, self.filename))
        self.group_name = group_name
        attrs = self.handle.get_analysis_attributes(group_name)

        if attrs is None:
            self.handle.add_analysis(self.analysis_id, group_name, meta, config)
            attrs = self.handle.get_analysis_attributes(group_name)
        if 'component' in attrs and attrs['component'] != self.analysis_id:
            raise ValueError('Component {} is not {}'.format(attrs.get('component'), self.analysis_id))

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()
        return False

    def close(self):
        """ Closes the object.
        """
        if self.handle and self.close_handle_when_done:
            self.handle.close()



import pkg_resources


def register_plugin():
    plugin_path = pkg_resources.resource_filename('ont_fast5_api', 'vbz_plugin')
    try:
        from h5py import h5pl
        h5pl.prepend(bytes(plugin_path, 'UTF-8'))
    except (ImportError, AttributeError):
        # We don't have the plugin library in h5py<2.10 so we fall back on an environment variable
        import os
        os.environ['HDF5_PLUGIN_PATH'] = plugin_path
    return plugin_path


class AbstractCompression:
    compression = "AbstractCompression"
    compression_opts = ()
    shuffle = False
    scaleoffset = False
    fletcher32 = False

    def __repr__(self):
        return self.compression

    @property
    def filter_settings(self):
        return {}


class VbzCompressionV1Alpha(AbstractCompression):
    def __init__(self):
        self.compression = 32020  # https://portal.hdfgroup.org/display/support/Registered+Filters
        self.compression_opts = (1, 2, 1, 1)  # VBZ_VERSION, VBZ_PACKING, VBZ_ZIG_ZAG, VBZ_ZSTD_COMPRESSION

    def __repr__(self):
        return "vbz_v1.alpha"

    @property
    def filter_settings(self):
        return {str(self.compression): self.compression_opts}


class VbzCompression(AbstractCompression):
    def __init__(self):
        self.compression = 32020  # https://portal.hdfgroup.org/display/support/Registered+Filters
        self.compression_opts = (0, 2, 1, 1)  # VBZ_VERSION, VBZ_PACKING, VBZ_ZIG_ZAG, VBZ_ZSTD_COMPRESSION

    def __repr__(self):
        return "vbz"

    @property
    def filter_settings(self):
        return {str(self.compression): self.compression_opts}


class GzipCompression(AbstractCompression):
    def __init__(self):
        self.compression = "gzip"
        self.compression_opts = 1

    @property
    def filter_settings(self):
        return {str(self.compression): self.compression_opts}


VBZ_ERROR_MESSAGE = "Failed to read compressed raw data. " \
                    "VBZ compression filter (id=32020) may be missing from expected path: '{}'"


def raise_missing_vbz_error_read(err):
    if str(VBZ.compression) in str(err):
        raise IOError(VBZ_ERROR_MESSAGE.format(register_plugin())) from err
    # If we don't see anything relating to VBZ just raise the existing error without additional info
    raise


def raise_missing_vbz_error_write(err):
    if type(err) is OSError and "Can't read data" in str(err):
        raise IOError(VBZ_ERROR_MESSAGE.format(register_plugin())) from err
    # If we don't see anything relating to VBZ just raise the existing error without additional info
    raise


VBZ_ALPHA = VbzCompressionV1Alpha()
VBZ = VbzCompression()
GZIP = GzipCompression()

COMPRESSION_MAP = {str(comp): comp for comp in (VBZ_ALPHA, VBZ, GZIP)}



""" Helper class for working with segmentation type analyses.
"""
import numpy as np

from ont_fast5_api.analysis_tools.base_tool import BaseTool
from ont_fast5_api.analysis_tools.event_detection import EventDetectionTools


class SegmentationTools(BaseTool):
    """ Provides helper methods specific to segmentation analyses.
    """
    group_id = 'Segmentation'
    analysis_id = 'segmentation'

    def get_results(self):
        """ Returns the segmentation summary data.
        
        This data is normalized, to eliminate differences in what is stored
        for different types of segmentation analyses.
        
        The following fields are output:

        * has_template - True if the segmentation found template data.
        * has_complement - True if the segmentation found complement data.
        * first_sample_template - The first sample of the template data in
            the raw data. Only present if has_template is True.
        * duration_template - The duration (in samples) of the template
            data. Only present if has_template is True.
        * first_sample_complement - The first sample of the complement data
            in the raw data. Only present if has_complement is True.
        * duration_complement - The duration (in samples) of the complement
            data. Only present if has_complement is True.
            
        """
        summary = self._get_summary_data()
        if summary is None:
            results = {'has_template': False,
                       'has_complement': False}
        else:
            results = {}
            if 'has_template' in summary:
                results['has_template'] = bool(summary['has_template'])
            else:
                results['has_template'] = True if summary['num_temp'] > 0 else False
            if 'has_complement' in summary:
                results['has_complement'] = bool(summary['has_complement'])
            else:
                results['has_complement'] = True if summary['num_comp'] > 0 else False
            need_raw_info = False
            if results['has_template']:
                if 'start_index_temp' in summary:
                    summary['start_event_template'] = summary['start_index_temp']
                    summary['end_event_template'] = summary['end_index_temp']
                if 'first_sample_template' not in summary:
                    need_raw_info = True
            if results['has_complement']:
                if 'start_index_comp' in summary:
                    summary['start_event_complement'] = summary['start_index_comp']
                    summary['end_event_complement'] = summary['end_index_comp']
                if 'first_sample_complement' not in summary:
                    need_raw_info = True
            if need_raw_info:
                self._get_raw_info(summary)
            if results['has_template']:
                results['first_sample_template'] = summary['first_sample_template']
                results['duration_template'] = summary['duration_template']
                if 'start_event_template' in summary:
                    results['start_event_template'] = summary['start_event_template']
                    results['end_event_template'] = summary['end_event_template']
            if results['has_complement']:
                results['first_sample_complement'] = summary['first_sample_complement']
                results['duration_complement'] = summary['duration_complement']
                if 'start_event_complement' in summary:
                    results['start_event_complement'] = summary['start_event_complement']
                    results['end_event_complement'] = summary['end_event_complement']
        return results

    def get_event_data(self, section, time_in_seconds=False):
        """ Get the template or complement event data.
        
        :param section: Either template, complement, or both.
        :param time_in_seconds: Return the start and length fields
            in seconds, rather than samples.
        :return: The event dataset for the section. If section=both
            then it returns a tuple with both sections. Returns None
            if the section does not exist.
        """
        if section not in ['template', 'complement', 'both']:
            raise Exception('Unrecognized section: {} Expected: "template", "complement" or "both"'.format(section))
        results = self.get_results()
        if results is None:
            return None, None if section is 'both' else None
        if section == 'both':
            sections = ['template', 'complement']
        else:
            sections = [section]
        evdet_group, _ = self._find_event_data()
        with EventDetectionTools(self.handle, group_name=evdet_group) as evdet:
            event_data, _ = evdet.get_event_data(time_in_seconds=time_in_seconds)
        datasets = [None, None]
        for n, this_section in enumerate(sections):
            if not results['has_{}'.format(this_section)]:
                continue
            ev1 = results['start_event_{}'.format(this_section)]
            ev2 = results['end_event_{}'.format(this_section)]
            datasets[n] = event_data[ev1:ev2]
        if section == 'both':
            return tuple(datasets)
        return datasets[0]

    def get_raw_data(self, section, scale=False):
        """ Get the template or complement raw data.
        
        :param section: Either template, complement, or both.
        :param scale: Scale the raw data to pA.
        :return:  The raw data for the section. If section=both
            then it returns a tuple with both sections. Returns None
            if the section does not exist.
        """
        results = self.get_results()
        datasets = [None, None]
        if section == 'both':
            sections = ['template', 'complement']
        else:
            sections = [section]
        for n, this_section in enumerate(sections):
            if not results['has_{}'.format(this_section)]:
                continue
            start = results['first_sample_{}'.format(this_section)]
            dur = results['duration_{}'.format(this_section)]
            datasets[n] = self.handle.get_raw_data(start=start, end=start+dur, scale=scale)
        if section == 'both':
            return tuple(datasets)
        return datasets[0]


    ##########################
    #
    #  Private methods below
    #
    ##########################
    
    def _get_summary_data(self):
        summary = self.handle.get_summary_data(self.group_name)
        if summary is None:
            return None
        if 'segmentation' in summary:
            results = summary['segmentation']
        elif 'split_hairpin' in summary:
            results = summary['split_hairpin']
        else:
            results = None
        return results

    def _find_event_data(self):
        attrs = self.handle.get_analysis_attributes(self.group_name)
        evdet_group = attrs.get('event_detection')
        if evdet_group is None:
            evdet_group = self.handle.get_latest_analysis('EventDetection')
        else:
            evdet_group = evdet_group[9:]
        if evdet_group is None:
            return None
        # We directly use the Fast5Read interface here, rather than the
        # EventDetectionTools one, because we don't want to load the entire
        # event table into memory.
        read_info = self.handle.status.read_info[0] # We assume only one read.
        read_number = read_info.read_number
        event_table_group = '{}/Reads/Read_{}'.format(evdet_group, read_number)
        dataset = self.handle.get_analysis_dataset(event_table_group, 'Events', skip_decoding=True)
        return evdet_group, dataset

    def _get_raw_info(self, summary):
        _, dataset = self._find_event_data()
        read_info = self.handle.status.read_info[0] # We assume only one read.
        if dataset is None:
            summary['first_sample_template'] = None
            summary['duration_template'] = None
            summary['first_sample_complement'] = None
            summary['duration_complement'] = None
            return
        if summary.get('start_event_template', -1) >= 0:
            ev1 = summary['start_event_template']
            ev2 = summary['end_event_template']
            summary['first_sample_template'] = dataset[ev1]['start'] - read_info.start_time
            end = dataset[ev2-1]['start'] + dataset[ev2-1]['length'] - read_info.start_time
            summary['duration_template'] = end - summary['first_sample_template']
        if summary.get('start_event_complement', -1) >= 0:
            ev1 = summary['start_event_complement']
            ev2 = summary['end_event_complement']
            summary['first_sample_complement'] = dataset[ev1]['start'] - read_info.start_time
            end = dataset[ev2-1]['start'] + dataset[ev2-1]['length'] - read_info.start_time
            summary['duration_complement'] = end - summary['first_sample_complement']



""" Helper class for working with 1D basecall type analyses.
"""
import numpy as np

from ont_fast5_api.analysis_tools.base_tool import BaseTool


class Basecall1DTools(BaseTool):
    """ Provides helper methods specific to 1D basecall analyses.
    """
    group_id = 'Basecall_1D'
    analysis_id = 'basecall_1d'


    def get_event_data(self, section):
        """ Return either the template or complement event data, if present.
        
        :param section: Either template or complement.
        :return: Event data table.
        """
        event_group = '{}/BaseCalled_{}'.format(self.group_name, section)
        data = self.handle.get_analysis_dataset(event_group, 'Events')
        return data

    def add_event_data(self, section, data):
        """ Add template or complement basecalled event data.
        
        :param section: Either template or complement.
        :param data: Event data table to be written.
        """
        event_group = 'BaseCalled_{}'.format(section)
        if not event_group in self.handle.handle['Analyses/{}'.format(self.group_name)]:
            self.handle.add_analysis_subgroup(self.group_name, event_group)
        self.handle.add_analysis_dataset('{}/{}'.format(self.group_name, event_group), 'Events', data)

    def get_called_sequence(self, section, fastq=False):
        """ Return either the called sequence data, if present.
        
        :param section: ['template', 'complement' or '2D']
        :param fastq: If True, return a single, multiline fastq string. If
            False, return a tuple of (name, sequence, qstring).
        :return: Either the fastq string or the (name, sequence, qstring) tuple.
        """

        event_group = '{}/BaseCalled_{}'.format(self.group_name, section)
        data = self.handle.get_analysis_dataset(event_group, 'Fastq')
        if data is None:
            raise KeyError("No fastq data in: {} {}".format(event_group, self.filename))
        if fastq:
            return data
        name, sequence, _, qstring = data.strip().split('\n')
        name = name[1:]
        return name, sequence, qstring

    def add_called_sequence(self, section, name, sequence, qstring):
        """ Add basecalled sequence data
        
        :param section: ['template', 'complement' or '2D']
        :param name: The record ID to use for the fastq.
        :param sequence: The called sequence.
        :param qstring: The quality string.
        """
        event_group = 'BaseCalled_{}'.format(section)
        if not event_group in self.handle.handle['Analyses/{}'.format(self.group_name)]:
            self.handle.add_analysis_subgroup(self.group_name, event_group)
        fastq_text = '@{}\n{}\n+\n{}\n'.format(name, sequence, qstring)
        fastq_arr = np.array(fastq_text, dtype=str)
        self.handle.add_analysis_dataset('{}/{}'.format(self.group_name, event_group), 'Fastq', fastq_arr)



""" Helper class for working with event detection type analyses.
"""
import numpy as np

from ont_fast5_api.analysis_tools.base_tool import BaseTool


class EventDetectionTools(BaseTool):
    """ Provides helper methods specific to event detection analyses.
    """

    group_id = 'EventDetection'
    analysis_id = 'event_detection'

    def set_event_data(self, data, read_attrs):
        """ Set event data with the specied attributes.
        
        :param data: Event data table.
        :param read_attrs: Attributes to put on the read group. This must include
            the read_number, which must refer to a read present in the object. The
            attributes should not include the standard read attributes:

             * read_id
             * start_time
             * duration
             * start_mux

            Those will be pulled from the read information already present in the
            object for the specified read.
        """
        if self.handle.mode == 'r':
            raise Exception('File is not open for writing.')
        read_number = read_attrs['read_number']
        read_group = '{}/Reads/Read_{}'.format(self.group_name, read_number)
        read_info = self.handle.status.read_info
        read_number_map = self.handle.status.read_number_map
        index = read_number_map.get(read_number)
        if index is None:
            raise Exception('Cannot add event detection data for a read that does not exist.')
        info = read_info[index]
        read_attrs.update({'read_id': info.read_id,
                           'start_time': info.start_time,
                           'duration': info.duration,
                           'start_mux': info.start_mux,
                           'median_before': info.median_before})
        attrs = self.handle.get_analysis_attributes(read_group)
        if attrs is None:
            self.handle.add_analysis_subgroup(self.group_name, 'Reads/Read_{}'.format(read_number),
                                              attrs=read_attrs)
            self.handle.add_analysis_dataset(read_group, 'Events', data)
        else:
            raise Exception('Event detection data already exists for this analysis and read.')

    def get_event_data(self, read_number=None, time_in_seconds=False):
        """ Get event data for the specified (or only) read.
        
        :param read_number: The read number to grab event data for. If this
            is None, and there is only one read, it will grab event data for
            that read.
        :param time_in_seconds: If True, this will convert (if necessary) the
            start and length fields from samples to seconds. If they are already
            in seconds, this option has no effect.
        :return: A tuple containing the event data, and the read attributes.
        """
        read_info = self.handle.status.read_info
        if read_number is None:
            if len(read_info) != 1:
                raise Exception('Must specify a read number if there is not exactly 1 read.')
            read_number = read_info[0].read_number
        else:
            read_numbers = [info.read_number for info in read_info]
            if read_number not in read_numbers:
                raise Exception('Specified read does not exist.')
        group = '{}/Reads/Read_{}'.format(self.group_name, read_number)
        attrs = self.handle.get_analysis_attributes(group)
        dataset = self.handle.get_analysis_dataset(group, 'Events', skip_decoding=True)
        if dataset is None:
            raise Exception('Read number {} has no event data.'.format(read_number))
        if time_in_seconds and dataset['start'].dtype.kind in ['i', 'u']:
            channel_info = self.handle.get_channel_info()
            sample_size = 1.0 / channel_info['sampling_rate']
            descr = [(x[0], 'float64') if x[0] in ('start', 'length') else x
                     for x in dataset.dtype.descr]
            data = dataset.astype(np.dtype(descr))[()]
            data['start'] *= sample_size
            data['length'] *= sample_size
        else:
            data = dataset[()]
        return data, attrs

    def has_event_data(self, read_number=None):
        """ Find out if the specified (or only) read has event data.

        :param read_number: The read number to check for event data. If this
            is ``None``, and there is only one read, it will check that read.
        :returns: True if event data exists for the read number.
        """
        read_info = self.handle.status.read_info
        if read_number is None:
            if len(read_info) != 1:
                raise Exception('Must specify a read number if there is not exactly 1 read.')
            read_number = read_info[0].read_number
        else:
            read_numbers = [info.read_number for info in read_info]
            if read_number not in read_numbers:
                raise Exception('Specified read does not exist.')
        group = '{}/Reads/Read_{}'.format(self.group_name, read_number)
        dataset = self.handle.get_analysis_dataset(group, 'Events', skip_decoding=True)
        return dataset is not None

    ##########################
    #
    #  Private methods below
    #
    ##########################

    def _new_analysis(self, meta, config):
        if self.handle.mode == 'r':
            raise Exception('Cannot create new event detection group. File is not open for writing.')
        self.handle.add_analysis('event_detection', self.group_name, meta, config)
        self.handle.add_analysis_subgroup(self.group_name, 'Reads')



import h5py
import warnings
from ont_fast5_api import fast5_file as f5f


class FileToDict(object):

    def __init__(self):
        self.contents = {}

    def scan(self, name, obj):
        if isinstance(obj, h5py.Group):
            self.contents[name] = 'group'
        else:
            self.contents[name] = 'dataset'
            self.contents['{}.data'.format(name)] = \
                str(f5f._sanitize_data_for_reading(obj))
            self.contents['{}.cols'.format(name)] = obj.dtype.names
        attrdict = {}
        for item in obj.attrs:
            attrdict[item] = \
                str(f5f._sanitize_data_for_reading(obj.attrs[item]))
        self.contents['{}.attrs'.format(name)] = attrdict
        return None


def compare_hdf_files(file1, file2):
    """ Compare two hdf files.
    :param file1: First file to compare.
    :param file2: Second file to compare.

    :returns True if they are the same.
    """
    warnings.simplefilter("default")
    warnings.warn("'compare_hdf_files': HDF5 comparison is deprecated. \n"
                  "If this feature is still required please contact the project maintainers",
                  DeprecationWarning)
    data1 = FileToDict()
    data2 = FileToDict()
    scanner1 = data1.scan
    scanner2 = data2.scan
    with h5py.File(file1, 'r') as fh1:
        fh1.visititems(scanner1)
    with h5py.File(file2, 'r') as fh2:
        fh2.visititems(scanner2)
    return data1.contents == data2.contents



import numpy
import os
import random

from ont_fast5_api.fast5_file import Fast5File
from ont_fast5_api.fast5_read import Fast5Read
from ont_fast5_api.multi_fast5 import MultiFast5File
from test.helpers import TestFast5ApiHelper

hexdigits = "0123456789abcdef"
run_id = "123abc"


class TestMultiFast5(TestFast5ApiHelper):

    def create_multi_file(self, read_ids):
        filename = self.generate_temp_filename()
        # driver=None is the default, but adding this in here makes sure we
        # preserve the constructor argument.
        with MultiFast5File(filename, 'w', driver=None) as multi_f5:
            for read_id in read_ids:
                multi_f5.create_empty_read(read_id, run_id)
        return filename

    def test_read_interface(self):
        read_ids = generate_read_ids(6)
        f5_file = self.create_multi_file(read_ids)

        with MultiFast5File(f5_file, 'a') as multi_f5:
            # Check we have the read_ids we expect
            self.assertEqual(sorted(read_ids), sorted(multi_f5.get_read_ids()))

            # Try and add another read with the same read_id and expect error
            with self.assertRaises(ValueError):
                multi_f5.create_empty_read(read_ids[0], run_id)

            # Test we can get a read from the file and it has the interface we expect
            read_0 = multi_f5.get_read(read_ids[0])
            self.assertTrue(isinstance(read_0, Fast5Read))

            # Test we cannot get a read which doesn't exit
            with self.assertRaises(KeyError):
                multi_f5.get_read("0123")

    def test_raw_data(self):
        f5_file = self.create_multi_file(generate_read_ids(4))
        data = list(range(10))
        raw_attrs = {
            "duration": 1,
            "median_before": 2.5,
            "read_id": "abcd",
            "read_number": 8,
            "start_mux": 2,
            "start_time": 99
        }
        with MultiFast5File(f5_file, 'a') as multi_f5:
            read0 = multi_f5.get_read(multi_f5.get_read_ids()[0])
            read0.add_raw_data(data, attrs=raw_attrs)
            output_data = read0.get_raw_data()
            numpy.testing.assert_array_equal(output_data, data)

    def test_channel_info(self):
        f5_file = self.create_multi_file(generate_read_ids(4))
        channel_info = {
            "digitisation": 2048,
            "offset": -119.5,
            "range": 74.2,
            "sampling_rate": 4000,
            "channel_number": "72"
        }
        # Fast5File explicitly casts the channel number on reading
        expected_out = channel_info.copy()
        expected_out['channel_number'] = int(channel_info['channel_number'])
        with MultiFast5File(f5_file, 'a') as multi_f5:
            read0 = multi_f5.get_read(multi_f5.get_read_ids()[0])
            read0.add_channel_info(channel_info)
            output_data = read0.get_channel_info()
            self.assertEqual(output_data, expected_out)

    def test_tracking_id(self):
        f5_file = self.create_multi_file(generate_read_ids(4))
        tracking_id = {
            "asic_id_eeprom": "some string",
            "device_id": "some string",
            "exp_script_name": "some string",
            "exp_script_purpose": "some string",
            "exp_start_time": "some string",
            "flow_cell_id": "some string",
            "hostname": "some string",
            "protocol_run_id": "some string",
            "protocols_version": "some string",
            "run_id": "some string",
            "version": "some string",
        }

        with MultiFast5File(f5_file, 'a') as multi_f5:
            read0 = multi_f5.get_read(multi_f5.get_read_ids()[0])
            read0.add_tracking_id(tracking_id)
            output_data = read0.get_tracking_id()
            self.assertEqual(output_data, tracking_id)

    def test_add_analysis(self):
        f5_file = self.create_multi_file(generate_read_ids(4))
        group = "Test"
        component = "test_component"
        attrs = {"attribute": 1}

        # Fast5File.add_analysis includes the component name in the analysis attributes
        expected_attributes = attrs.copy()
        expected_attributes['component'] = component
        with MultiFast5File(f5_file, 'a') as multi_f5:
            read0 = multi_f5.get_read(multi_f5.get_read_ids()[0])
            self.assertEqual(read0.list_analyses(), [])
            read0.add_analysis(component, group, attrs)
            self.assertEqual(read0.list_analyses(), [(component, group)])
            self.assertEqual(read0.get_analysis_attributes(group), expected_attributes)


def generate_read_ids(num_ids, id_len=8):
    return ["".join(random.choice(hexdigits) for _ in range(id_len)) for _ in range(num_ids)]



import numpy as np
from ont_fast5_api.fast5_file import Fast5File
from ont_fast5_api.analysis_tools.basecall_1d import Basecall1DTools
from test.helpers import TestFast5ApiHelper


class TestBasecall1DTools(TestFast5ApiHelper):

    def test_001_put_and_retrieve(self):
        fname = self.generate_temp_filename()
        dtypes = [('mean', float),
                  ('start', float),
                  ('stdv', float),
                  ('length', float),
                  ('called_state', '<U5'),
                  ('move', int)]
        data1 = np.zeros(10, dtype=dtypes)
        data1['mean'] = [10.0, 15.0, 8.5, 7.2, 13.6, 9.4, 11.8, 10.1, 4.2, 10.9]
        data1['stdv'] = [0.7, 0.9, 1.0, 1.1, 0.75, 0.6, 0.83, 1.12, 9.45, 2.9]
        data1['start'] = [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
        data1['length'] = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]
        data1['move'] = [1, 1, 0, 1, 0, 2, 1, 1, 1, 0]
        data1['called_state'] = ['AAAAA', 'AAAAT', 'AAAAT', 'AAATC', 'AAATC',
                                 'ATCCG', 'TCCGT', 'CCGTT', 'CGTTA', 'CGTTA']
        data2 = data1[::-1]
        seq1 = 'AAAAATCCGTTA'
        seq2 = 'TAACGGATTTTT'
        qstring1 = 'blahblahblah'
        qstring2 = 'halbhalbhalb'
        with Fast5File(fname, mode='w') as fh:
            fh.add_channel_info({'channel_number': 1,
                                 'sampling_rate': 4000,
                                 'digitisation': 8192,
                                 'range': 819.2,
                                 'offset': 0})
            fh.add_read(12, 'unique_snowflake', 12345, 1000, 0, 120.75)
            attrs = {'name': 'test', 'version': 0, 'time_stamp': 'just now'}
            fh.add_analysis('basecall_1d', 'Basecall_1D_000', attrs)
            with Basecall1DTools(fh, group_name='Basecall_1D_000') as basecall:
                basecall.add_event_data('template', data1)
                basecall.add_event_data('complement', data2)
                basecall.add_called_sequence('template', 'template', seq1, qstring1)
                basecall.add_called_sequence('complement', 'complement', seq2, qstring2)
        with Fast5File(fname, mode='r') as fh:
            with Basecall1DTools(fh, group_name='Basecall_1D_000') as basecall:
                events1 = basecall.get_event_data('template')
                np.testing.assert_array_equal(events1, data1)
                events2 = basecall.get_event_data('complement')
                np.testing.assert_array_equal(events2, data2)
                n1, s1, q1 = basecall.get_called_sequence('template')
                self.assertEqual(n1, 'template')
                self.assertEqual(s1, seq1)
                self.assertEqual(q1, qstring1)
                n2, s2, q2 = basecall.get_called_sequence('complement')
                self.assertEqual(n2, 'complement')
                self.assertEqual(s2, seq2)
                self.assertEqual(q2, qstring2)



""" Helper class for working with 2D basecall type analyses.
"""
import warnings
from ont_fast5_api.analysis_tools.basecall_1d import Basecall1DTools


class Basecall2DTools(Basecall1DTools):
    """ Provides helper methods specific to 2D basecall analyses.
    """

    group_id = 'Basecall_2D'
    analysis_id = 'basecall_2d'

    def get_prior_alignment(self):
        """ Return the prior alignment that was used for 2D basecalling.

        :return: Alignment data table.
        """
        data_group = '{}/HairpinAlign'.format(self.group_name)
        data = self.handle.get_analysis_dataset(data_group, 'Alignment')
        return data

    def get_2d_call_alignment(self):
        """ Return the alignment and model_states from the 2D basecall.

        :return: Alignment data table.
        """
        data_group = '{}/BaseCalled_2D'.format(self.group_name)
        data = self.handle.get_analysis_dataset(data_group, 'Alignment')
        return data

    def add_prior_alignment(self, data):
        """ Add template or complement basecalled event data.
        
        :param data: Alignment table to be written.
        """
        path = 'Analyses/{}'.format(self.group_name)
        if 'HairpinAlign' not in self.handle.handle[path]:
            self.handle.add_analysis_subgroup(self.group_name, 'HairpinAlign')

        path = '{}/HairpinAlign'.format(self.group_name)
        self.handle.add_analysis_dataset(path, 'Alignment', data)

    def add_2d_call_alignment(self, data):
        """ Add the alignment and model_state data table..
        
        :param data: Alignment and model_state table to be written.
        """
        path = 'Analyses/{}'.format(self.group_name)
        if 'BaseCalled_2D' not in self.handle.handle[path]:
            self.handle.add_analysis_subgroup(self.group_name, 'BaseCalled_2D')

        path = '{}/BaseCalled_2D'.format(self.group_name)
        self.handle.add_analysis_dataset(path, 'Alignment', data)

    def get_called_sequence(self, section=None, fastq=False):
        """ Return either the called sequence data, if present.
        :param section: ['template', 'complement' or '2D']
        :param fastq: If True, return a single, multiline fastq string. If
            False, return a tuple of (name, sequence, qstring).
        :return: Either the fastq string or the (name, sequence, qstring) tuple.
        """
        if section != "2D":
            warnings.warn("Basecall2DTools.get_called_sequence() should specify section='2D'", DeprecationWarning)
            # Backwards compatibilty to 0.3.3, if no "2D" section, bump args by 1 and pass to super
            if section == None:
                # We assume that a named arg or no-arg was given
                return super(Basecall2DTools, self).get_called_sequence("2D", fastq)
            # We assume that a single unnamed arg was given for fastq
            return super(Basecall2DTools, self).get_called_sequence("2D", section)
        return super(Basecall2DTools, self).get_called_sequence(section, fastq)
